from roux.global_imports import *

# curate data 
def drop_low_complexity(df1: pd.DataFrame,
                        min_nunique: int,
                        max_inflation: int,
                        cols: list=None,
                        cols_keep: list=[],
                        test: bool=False) -> pd.DataFrame:
    """Remove low-complexity columns from the data. 

    Args:
        df1 (pd.DataFrame): input data.
        min_nunique (int): minimum unique values.
        max_inflation (int): maximum over-representation of the values.
        cols (list, optional): columns. Defaults to None.
        cols_keep (list, optional): columns to keep. Defaults to [].
        test (bool, optional): test mode. Defaults to False.

    Returns:
        pd.DataFrame: output data.
    """

    if cols is None:
        cols=df1.columns.tolist()
    df_=pd.concat([df1.rd.check_nunique(cols),df1.rd.check_inflation(cols)],axis=1,)
    df_.columns=['nunique','% inflation']
    df_=df_.sort_values(df_.columns.tolist(),ascending=False)
    df_=df_.loc[((df_['nunique']<=min_nunique) | (df_['% inflation']>=max_inflation)),:]
    l1=df_.index.tolist()
#     def apply_(x,df1,min_nunique,max_inflation):
#         ds1=x.value_counts()
#         return (len(ds1)<=min_nunique) or ((ds1.values[0]/len(df1))>=max_inflation)
#     l1=df1.loc[:,cols].apply(lambda x: apply_(x,df1,min_nunique=min_nunique,max_inflation=max_inflation)).loc[lambda x: x].index.tolist()
    logging.info(f"{len(l1)}(/{len(cols)}) low complexity columns {'could be ' if test else ''}dropped:")
    info(df_)
    if len(cols_keep)!=0:
        assert all([c in df1 for c in cols_keep]), ([c for c in cols_keep if not c in df1])
        cols_kept=[c for c in l1 if c in cols_keep]
        info(cols_kept)
        l1=[c for c in l1 if not c in cols_keep]
    if not test:
        return df1.log.drop(labels=l1,axis=1)
    else:
        return df1

def get_Xy_for_classification(df1: pd.DataFrame,coly: str,qcut: float=None,
                              # low_complexity filters
                              drop_xs_low_complexity: bool=False,
                              min_nunique: int=5,
                              max_inflation: float=0.5,
                              **kws,
                             ) -> dict:
    """Get X matrix and y vector. 
    
    Args:
        df1 (pd.DataFrame): input data, should be indexed.
        coly (str): column with y values, bool if qcut is None else float/int
        qcut (float, optional): quantile cut-off. Defaults to None.
        drop_xs_low_complexity (bool, optional): to drop columns with <5 unique values. Defaults to False.
        min_nunique (int, optional): minimum unique values in the column. Defaults to 5.
        max_inflation (float, optional): maximum inflation. Defaults to 0.5.

    Keyword arguments:
        kws: parameters provided to `drop_low_complexity`.

    Returns:
        dict: output.
    """
    df1=df1.rd.clean(drop_constants=True)
    cols_X=[c for c in df1 if c!=coly]
    if not qcut is None:
        if qcut>0.5:
            logging.error('qcut should be <=0.5')
            return 
        lims=[df1[coly].quantile(1-qcut),df1[coly].quantile(qcut)]
        df1[coly]=df1.progress_apply(lambda x: True if x[coly]>=lims[0] else False if x[coly]<lims[1] else np.nan,axis=1)
        df1=df1.log.dropna()
    df1[coly]=df1[coly].apply(bool)
    info(df1[coly].value_counts())
    y=df1[coly]
    X=df1.loc[:,cols_X]
    # remove low complexity features
    X=X.rd.clean(drop_constants=True)
    X=drop_low_complexity(X,cols=None,
                          min_nunique=min_nunique,
                          max_inflation=max_inflation,
                          test=False if drop_xs_low_complexity else True,
                          **kws,
                         )
    return {'X':X,'y':y}

def get_cvsplits(
    X: np.array,
    y: np.array,
    cv: int=5,
    random_state: int=None,
    outtest: bool=True
    ) -> dict:
    """Get cross-validation splits.

    Args:
        X (np.array): X matrix.
        y (np.array): y vector.
        cv (int, optional): cross validations. Defaults to 5.
        random_state (int, optional): random state. Defaults to None.
        outtest (bool, optional): output testing. Defaults to True.

    Returns:
        dict: output.
    """
    if random_state is None: logging.warning(f"random_state is None")
    X.index=range(len(X))
    y.index=range(len(y))
    
    from sklearn.model_selection import KFold
    cv = KFold(n_splits=cv,random_state=random_state,shuffle=True)
    cv2Xy={}
    for i, (train ,test) in enumerate(cv.split(X.index)):
        dtype2index=dict(zip(('train' ,'test'),(train ,test)))
        cv2Xy[i]={}
        if outtest:
            for dtype in dtype2index:
                cv2Xy[i][dtype]={}
                cv2Xy[i][dtype]['X' if isinstance(X,pd.DataFrame) else 'x']=X.iloc[dtype2index[dtype],:] if isinstance(X,pd.DataFrame) else X[dtype2index[dtype]]
                cv2Xy[i][dtype]['y']=y[dtype2index[dtype]]
        else:
            cv2Xy[i]['X' if isinstance(X,pd.DataFrame) else 'x']=X.iloc[dtype2index['train'],:] if isinstance(X,pd.DataFrame) else X[dtype2index['train']]
            cv2Xy[i]['y']=y[dtype2index['train']]                
    return cv2Xy

# search estimator
def get_grid_search(modeln: str,
                    X: np.array,
                    y: np.array,
                    param_grid: dict={},
                    cv: int=5,
                    n_jobs: int=6,
                    random_state: int=None,
                    scoring: str='balanced_accuracy',
                    **kws,
                   ) -> object:
    """Grid search.

    Args:
        modeln (str): name of the model.
        X (np.array): X matrix.
        y (np.array): y vector.
        param_grid (dict, optional): parameter grid. Defaults to {}.
        cv (int, optional): cross-validations. Defaults to 5.
        n_jobs (int, optional): number of cores. Defaults to 6.
        random_state (int, optional): random state. Defaults to None.
        scoring (str, optional): scoring system. Defaults to 'balanced_accuracy'.

    Keyword arguments:
        kws: parameters provided to the `GridSearchCV` function.

    Returns:
        object: `grid_search`.

    References: 
        1. https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html
        2. https://scikit-learn.org/stable/modules/model_evaluation.html
    """
    if random_state is None: logging.warning(f"random_state is None")
    from sklearn.model_selection import GridSearchCV
    from sklearn import ensemble
    estimator = getattr(ensemble,modeln)(random_state=random_state)
    grid_search = GridSearchCV(estimator, 
                               param_grid,
                               cv=cv,
                               n_jobs=n_jobs,
                               scoring=scoring,
                               **kws)
    grid_search.fit(X, y)
    info(modeln,grid_search.best_params_)
    info(modeln,grid_search.best_score_)
    return grid_search

def get_estimatorn2grid_search(estimatorn2param_grid: dict,
                                X: pd.DataFrame,
                                y: pd.Series,
                                **kws
                                ) -> dict:
    """Estimator-wise grid search.

    Args:
        estimatorn2param_grid (dict): estimator name to the grid search map.
        X (pd.DataFrame): X matrix.
        y (pd.Series): y vector.

    Returns:
        dict: output.
    """
    estimatorn2grid_search={}
    for k in tqdm(estimatorn2param_grid.keys()):
        estimatorn2grid_search[k]=get_grid_search(modeln=k,
                        X=X,y=y,
                        param_grid=estimatorn2param_grid[k],
                        cv=5,
                        n_jobs=6,
                        **kws,
                       )
#     info({k:estimatorn2grid_search[k].best_params_ for k in estimatorn2grid_search})
    return estimatorn2grid_search

def get_test_scores(d1: dict) -> pd.DataFrame:
    """Test scores.

    Args:
        d1 (dict): dictionary with objects.

    Returns:
        pd.DataFrame: output.

    TODOs: 
        Get best param index.
    """
    d2={}
    for k1 in d1:
#             info(k1,dict2str(d1[k1].best_params_))
        l1=list(d1[k1].cv_results_.keys())
        l1=[k2 for k2 in l1 if not re.match("^split[0-9]_test_.*",k2) is None]
        d2[k1+"\n("+dict2str(d1[k1].best_params_,sep='\n')+")"]={k2: d1[k1].cv_results_[k2] for k2 in l1}
    df1=pd.DataFrame(d2).applymap(lambda x: x[0] if (len(x)==1) else max(x)).reset_index()
    df1['variable']=df1['index'].str.split('_test_',expand=True)[1].str.replace('_',' ')
    df1['cv #']=df1['index'].str.split('_test_',expand=True)[0].str.replace('split','').apply(int)
    df1=df1.rd.clean()
    return df1.melt(id_vars=['variable','cv #'],
                   value_vars=d2.keys(),
                   var_name='model',
                   )

## evaluate metrics
def plot_metrics(outd: str,plot: bool=False) -> pd.DataFrame:
    """Plot performance metrics.

    Args:
        outd (str): output directory.
        plot (bool, optional): make plots. Defaults to False.

    Returns:
        pd.DataFrame: output data.
    """
    d0=read_dict(f'{outd}/input.json')
    d1=read_pickle(f'{outd}/estimatorn2grid_search.pickle')
    df01=read_table(f'{outd}/input.pqt')
    df2=get_test_scores(d1)
    df2.loc[(df2['variable']=='average precision'),'value reference']=sum(df01[d0['coly']])/len(df01[d0['coly']])
    if plot:
        _,ax=plt.subplots(figsize=[3,3])
        sns.pointplot(data=df2,
        y='variable',
        x='value',
        hue='model',
        join=False,
        dodge=0.2,
        ax=ax)
        ax.axvline(0.5,linestyle=":",
                   color='k',
                  label='reference: accuracy')
        ax.axvline(sum(df01[d0['coly']])/len(df01[d0['coly']]),linestyle=":",
                   color='b',
                  label='reference: precision')
        ax.legend(bbox_to_anchor=[1,1])
        ax.set(xlim=[-0.1,1.1])
    return df2

def get_probability(estimatorn2grid_search: dict,
                    X: np.array,y: np.array,
                    colindex: str,
                    coff: float=0.5,
                   test: bool=False):
    """Classification probability.

    Args:
        estimatorn2grid_search (dict): estimator to the grid search map.
        X (np.array): X matrix.
        y (np.array): y vector.
        colindex (str): index column. 
        coff (float, optional): cut-off. Defaults to 0.5.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        pd.DataFrame: output.
    """
    assert(all(X.index==y.index))
    df0=y.to_frame('actual').reset_index()
    df1=pd.DataFrame({k:estimatorn2grid_search[k].best_estimator_.predict(X) for k in estimatorn2grid_search})#.add_prefix('prediction ')
    df1.index=X.index
    df1=df1.melt(ignore_index=False,
            var_name='estimator',
            value_name='prediction').reset_index()
    df2=pd.DataFrame({k:estimatorn2grid_search[k].best_estimator_.predict_proba(X)[:,1] for k in estimatorn2grid_search})#.add_prefix('prediction probability ')
    df2.index=X.index
    df2=df2.melt(ignore_index=False,
            var_name='estimator',
            value_name='prediction probability').reset_index()

    df3=df1.log.merge(right=df2,
                  on=['estimator',colindex],
                 how='inner',
                 validate="1:1")\
           .log.merge(right=df0,
                  on=[colindex],
                 how='inner',
    #              validate="1:1",
                )
    ## predicted correctly
    df3['TP']=df3.loc[:,['prediction','actual']].all(axis=1)
    if test:
        def plot_(df5):
            assert len(df5)==4, df5
            df6=df5.pivot(index='prediction',columns='actual',values='count').sort_index(axis=0,ascending=False).sort_index(axis=1,ascending=False)
            from roux.viz.heatmap import plot_crosstab
            ax=plot_crosstab(df6,annot_pval=None,
                            confusion=True)
            ax.set_title(df5.name,loc='left')
        df4=df3.groupby('estimator').apply(lambda df: pd.crosstab(df['prediction'],df['actual']).melt(ignore_index=False,value_name='count')).reset_index()
        df4.groupby('estimator').apply(plot_)
    return df3
#     df1=dellevelcol(pd.concat({k:pd.DataFrame({'sample name':X.index,
#                                               'true':y,
#                                               'probability':estimatorn2grid_search[k].best_estimator_.predict_proba(X)[:,1],}) for k in estimatorn2grid_search,
#                                               'prediction':estimatorn2grid_search[k].best_estimator_.predict(X) for k in estimatorn2grid_search,
#                               },
#                              axis=0,names=['estimator name'],
#                              ).reset_index())
#     info(df1.shape)
#     df1.loc[:,'correct by truth']=df1.apply(lambda x: ((x['true'] and x['probability']>coff) or (not x['true'] and x['probability']<1-coff)),axis=1)
#     info(df1.loc[:,'correct by truth'].sum())

#     df1['probability per class']=df1.apply(lambda x: np.nan if not x['correct by truth'] else 1-x['probability'] if x['probability']<0.5 else x['probability'],axis=1)
#     if test:
#         plt.figure(figsize=[4,4])
#         ax=plt.subplot()
#         df1.groupby('estimator name').apply(lambda df: df['probability'].hist(bins=50,label=df.name,histtype='step'))
#         ax.axvline(coff,label='cut off')
#         ax.set(xlim=[0.5,1])
#         ax.legend(loc=2)
#         _=ax.set(xlabel='probability',ylabel='count')

#     df1=df1.merge(df1.groupby(['sample name']).agg({'probability per class': lambda x: all([i>coff or i<1-coff for i in x])}).rename(columns={'probability per class':'correct by estimators'}).reset_index(),
#              on='sample name',how='left')

#     info('total samples\t',len(df1))
#     info(df1.groupby(['sample name']).agg({c:lambda x: any(x) for c in df1.filter(regex='^correct ')}).sum())
#     return df1

def run_grid_search(df: pd.DataFrame,
    colindex: str,
    coly: str,
    n_estimators: int,
    qcut: float=None,
    evaluations: list=['prediction','feature importances',
    'partial dependence',
    ],
    estimatorn2param_grid: dict=None,
    drop_xs_low_complexity: bool=False,
    min_nunique: int=5,
    max_inflation: float=0.5,      
    cols_keep: list=[],
    outp: str=None,
    test: bool=False,
    **kws, ## grid search
    ) -> dict:
    """Run grid search.

    Args:
        df (pd.DataFrame): input data.
        colindex (str): column with the index.
        coly (str): column with y values. Data type bool if qcut is None else float/int.
        n_estimators (int): number of estimators.
        qcut (float, optional): quantile cut-off. Defaults to None.
        evaluations (list, optional): evaluations types. Defaults to ['prediction','feature importances', 'partial dependence', ].
        estimatorn2param_grid (dict, optional): estimator to the parameter grid map. Defaults to None.
        drop_xs_low_complexity (bool, optional): drop the low complexity columns. Defaults to False.
        min_nunique (int, optional): minimum unique values allowed. Defaults to 5.
        max_inflation (float, optional): maximum inflation allowed. Defaults to 0.5.
        cols_keep (list, optional): columns to keep. Defaults to [].
        outp (str, optional): output path. Defaults to None.
        test (bool, optional): test mode. Defaults to False.

    Keyword arguments:
        kws: parameters provided to `get_estimatorn2grid_search`.

    Returns:
        dict: estimator to grid search map.
    """
    assert('random_state' in kws)
    if kws['random_state'] is None: logging.warning(f"random_state is None")
    
    if estimatorn2param_grid is None: 
        from sklearn import ensemble
        estimatorn2param_grid={k:getattr(ensemble,k)().get_params() for k in estimatorn2param_grid}
        if test=='estimatorn2param_grid':
            return estimatorn2param_grid
    #     info(estimatorn2param_grid)
        for k in estimatorn2param_grid:
            if 'n_estimators' not in estimatorn2param_grid[k]:
                estimatorn2param_grid[k]['n_estimators']=[n_estimators]
        if test: info(estimatorn2param_grid)
        d={}
        for k1 in estimatorn2param_grid:
            d[k1]={}
            for k2 in estimatorn2param_grid[k1]:
                if isinstance(estimatorn2param_grid[k1][k2],list):
                    d[k1][k2]=estimatorn2param_grid[k1][k2]
        estimatorn2param_grid=d
    if test: info(estimatorn2param_grid)
    params=get_Xy_for_classification(df.set_index(colindex),coly=coly,
                                    qcut=qcut,drop_xs_low_complexity=drop_xs_low_complexity,
                                    min_nunique=min_nunique,
                                    max_inflation=max_inflation,
                                    cols_keep=cols_keep,
                                    )
    dn2df={}
    dn2df['input']=params['X'].join(params['y'])
    estimatorn2grid_search=get_estimatorn2grid_search(estimatorn2param_grid,
                                                      X=params['X'],y=params['y'],
                                                      **kws)
#     to_dict({k:estimatorn2grid_search[k].cv_results_ for k in estimatorn2grid_search},
#            f'{outp}/estimatorn2grid_search_results.json')
    if not outp is None:
        to_dict(estimatorn2grid_search,f'{outp}/estimatorn2grid_search.pickle')
        to_dict(estimatorn2grid_search,f'{outp}/estimatorn2grid_search.joblib')
        d1={} # cols
        d1['colindex']=colindex
        d1['coly']=coly
        d1['cols_x']=dn2df['input'].filter(regex=f"^((?!({d1['colindex']}|{d1['coly']})).)*$").columns.tolist()
        d1['estimatorns']=list(estimatorn2param_grid.keys())
        d1['evaluations']=evaluations
        to_dict(d1,f'{outp}/input.json')
#     return estimatorn2grid_search
    ## interpret
    kws2={'random_state':kws['random_state']}
    if 'prediction' in evaluations:
        dn2df['prediction']=get_probability(estimatorn2grid_search,
                                            X=params['X'],y=params['y'],
                                            colindex=colindex,
#                                             coly=coly,
                                            test=True,
#                                             **kws2,
                                           )

    if 'feature importances' in evaluations:
        dn2df['feature importances']=get_feature_importances(estimatorn2grid_search,
                                X=params['X'],y=params['y'],
                                test=test,**kws2)
    if 'partial dependence' in evaluations:
        dn2df['partial dependence']=get_partial_dependence(estimatorn2grid_search,
                                X=params['X'],y=params['y'],
#                                                            **kws2,
                                                          )
    ## save data
    if not outp is None:
        for k in dn2df:
            if isinstance(dn2df[k],dict):
                dn2df[k]=pd.concat(dn2df[k],axis=0,names=['estimator name']).reset_index(0)
            if 'permutation #' in dn2df[k]:
                dn2df[k]['permutation #']=dn2df[k]['permutation #'].astype(int)
            to_table(dn2df[k],f'{outp}/{k}.pqt')
        df_=plot_metrics(outd=outp,plot=True)
        to_table(df_,f'{outp}/metrics.pqt')
    else:
        return estimatorn2grid_search

# interpret 
def plot_feature_predictive_power(df3: pd.DataFrame,
                                  ax: plt.Axes=None,
                                  figsize: list=[3,3],
                                  **kws) -> plt.Axes:
    """Plot feature-wise predictive power.

    Args:
        df3 (pd.DataFrame): input data.
        ax (plt.Axes, optional): axes object. Defaults to None.
        figsize (list, optional): figure size. Defaults to [3,3].

    Returns:
        plt.Axes: output.
    """

    df4=df3.rd.filter_rows({'variable':'ROC AUC'}).rd.groupby_sort_values(col_groupby='feature',
                     col_sortby='value',
                     ascending=False)
    if ax is None:
        _,ax=plt.subplots(figsize=figsize)
    sns.pointplot(data=df3,
                 y='feature',
                 x='value',
                 hue='variable',
                  order=df4['feature'].unique(),
                 join=False,
                  ax=ax,
                 )
    ax.legend(bbox_to_anchor=[1,1])
    ax.axvline(0.5, linestyle=':', color='lightgray')
    return ax

def get_feature_predictive_power(d0: dict,df01: pd.DataFrame,
                                n_splits: int=5, 
                                n_repeats: int=10,
                                random_state: int=None,
                                 plot: bool=False,
                                 drop_na: bool=False,
                               **kws) -> pd.DataFrame:
    """get_feature_predictive_power _summary_

    Notes: 
        x-values should be scale and sign agnostic.

    Args:
        d0 (dict): input dictionary.
        df01 (pd.DataFrame): input data, 
        n_splits (int, optional): number of splits. Defaults to 5.
        n_repeats (int, optional): number of repeats. Defaults to 10.
        random_state (int, optional): random state. Defaults to None.
        plot (bool, optional): plot. Defaults to False.
        drop_na (bool, optional): drop missing values. Defaults to False.

    Returns:
        pd.DataFrame: output data.
    """
    if random_state is None: logging.warning(f"random_state is None")
    from sklearn.metrics import average_precision_score,roc_auc_score
    from sklearn.model_selection import StratifiedKFold,RepeatedStratifiedKFold

    d2={}
    for colx in tqdm(d0['cols_x']):
        df1=df01.loc[:,[d0['coly'],colx]]
        if drop_na:
            df1=df1.dropna() 
        if df1[d0['coly']].nunique()==1: continue
        if sum(df1[d0['coly']]==True)<5: continue
        if sum(df1[d0['coly']]==False)<5: continue        
        # if perc(df1[d0['coly']])>90: continue
        # if perc(df1[d0['coly']])<10: continue
        if df1[colx].nunique()==1: continue
        cv = RepeatedStratifiedKFold(n_splits=n_splits, n_repeats=n_repeats, random_state=random_state,**kws)
        d1={i: ids for i,(_, ids) in enumerate(cv.split(df1[colx], df1[d0['coly']]))}
        df2=pd.DataFrame({'cv #':range(cv.get_n_splits())})
        if roc_auc_score(df1[d0['coly']], df1[colx])<roc_auc_score(df1[d0['coly']], -df1[colx]):
#             df1[d0['coly']]=~df1[d0['coly']]
            df1[colx]=-df1[colx]
        try:
            df2['ROC AUC']=df2['cv #'].apply(lambda x: roc_auc_score(df1.iloc[d1[x],:][d0['coly']],
                                                                     df1.iloc[d1[x],:][colx]))
            df2['average precision']=df2['cv #'].apply(lambda x: average_precision_score(df1.iloc[d1[x],:][d0['coly']],
                                                                                         df1.iloc[d1[x],:][colx]))
        except:
            print(df1)
        d2[colx]=df2.melt(id_vars='cv #',value_vars=['ROC AUC','average precision'])

    df3=pd.concat(d2,axis=0,names=['feature']).reset_index(0)
    if plot: plot_feature_predictive_power(df3)
    return df3

def get_feature_importances(estimatorn2grid_search: dict,
                            X: pd.DataFrame,y: pd.Series,
                            scoring: str='roc_auc',
                            n_repeats: int=20,
                            n_jobs: int=6,
                            random_state: int=None,
                            plot: bool=False,
                            test: bool=False,
                           **kws) -> pd.DataFrame:
    """Feature importances.

    Args:
        estimatorn2grid_search (dict): map between estimator name and grid search object. 
        X (pd.DataFrame): X matrix.
        y (pd.Series): y vector.
        scoring (str, optional): scoring type. Defaults to 'roc_auc'.
        n_repeats (int, optional): number of repeats. Defaults to 20.
        n_jobs (int, optional): number of cores. Defaults to 6.
        random_state (int, optional): random state. Defaults to None.
        plot (bool, optional): plot. Defaults to False.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        pd.DataFrame: output data.
    """
    if random_state is None: logging.warning(f"random_state is None")    
    def plot_(df,ax=None):
        if ax is None:
            fig,ax=plt.subplots(figsize=[4,(df['estimator name'].nunique()*0.5)+2])
        dplot=groupby_sort_values(df,
             col_groupby=['estimator name','feature'],
             col_sortby='importance rescaled',
             func='mean', ascending=False
            )
        dplot=dplot.loc[(dplot['importance']!=0),:]

        sns.pointplot(data=dplot,
              x='importance rescaled',
              y='feature',
              hue='estimator name',
             linestyles=' ',
              markers='o',
              alpha=0.1,
              dodge=True,
              scatter_kws = {'facecolors':'none'},
              ax=ax
             )
        return ax
    
    dn2df={}
    for k in tqdm(estimatorn2grid_search.keys()):
        from sklearn.inspection import permutation_importance
        r = permutation_importance(estimator=estimatorn2grid_search[k].best_estimator_, 
                                   X=X, y=y,
                                   scoring=scoring,
                                   n_repeats=n_repeats,
                                   n_jobs=n_jobs,
                                   random_state=random_state,
                                   **kws,
                                  )
        df=pd.DataFrame(r.importances)
        df['feature']=X.columns
        dn2df[k]=df.melt(id_vars=['feature'],value_vars=range(n_repeats),
            var_name='permutation #',
            value_name='importance',
           )
    df2=pd.concat(dn2df,axis=0,names=['estimator name']).reset_index(0)
    from roux.stat.transform import rescale
    def apply_(df):
        df['importance rescaled']=rescale(df['importance'])
        df['importance rank']=len(df)-df['importance'].rank()
        return df#.sort_values('importance rescaled',ascending=False)
    df3=df2.groupby(['estimator name','permutation #']).apply(apply_)
    if plot:
        plot_(df3)
    return df3

def get_partial_dependence(estimatorn2grid_search: dict,
                            X: pd.DataFrame,y: pd.Series,
                            ) -> pd.DataFrame:
    """Partial dependence.

    Args:
        estimatorn2grid_search (dict): map between estimator name and grid search object.
        X (pd.DataFrame): X matrix.
        y (pd.Series): y vector.

    Returns:
        pd.DataFrame: output data.
    """
    df3=pd.DataFrame({'feature #':range(len(X.columns)),
                     'feature name':X.columns})

    def apply_(featuren,featurei,estimatorn2grid_search):
        from sklearn.inspection import partial_dependence
        dn2df={}
        for k in estimatorn2grid_search:
            t=partial_dependence(estimator=estimatorn2grid_search[k].best_estimator_,
                                 X=X,
                                 features=[featurei],
                                 response_method='predict_proba',
                                 method='brute',
                                 percentiles=[0,1],
                                 grid_resolution=100,
                                 )
            dn2df[k]=pd.DataFrame({'probability':t[0][0],
                                    'feature value':t[1][0]})
        df1=pd.concat(dn2df,axis=0,names=['estimator name']).reset_index()
        df1['feature name']=featuren
        return df1.rd.clean()
    df4=df3.groupby('feature #',as_index=False).progress_apply(lambda df:apply_(featuren=df.iloc[0,:]['feature name'],
                                                                                featurei=df.iloc[0,:]['feature #'],
                                                                                estimatorn2grid_search=estimatorn2grid_search))
    
    return df4






import pandas as pd

def get_subgraphs(df1: pd.DataFrame,
                 source: str,
                 target: str) -> pd.DataFrame:
    """Subgraphs from the the edge list.

    Args:
        df1 (pd.DataFrame): input dataframe containing edge-list.
        source (str): source node.
        target (str): taget node.

    Returns:
        pd.DataFrame: output.
    """
    import networkx as nx
    g=nx.from_pandas_edgelist(df1,source=source,target=target)
    ug = g.to_undirected()
    sgs = nx.connected_components(ug)
    dn2df={}
    for sg in sgs:
        ns=sorted(sg)
        dn2df['--'.join(ns)]=pd.Series(ns)
    return pd.concat(dn2df,names=['subnetwork name']).reset_index().drop(['level_1'],axis=1).rename(columns={0:'node name'})



from roux.global_imports import *
from roux.viz.ax_ import *

def plot_table(
    df1: pd.DataFrame,
    xlabel: str=None,
    ylabel: str=None,
    annot: bool=True,
    cbar: bool=False,
    linecolor: str='k',
    linewidths: float=1,               
    cmap: str=None,
    sorty: bool=False,
    linebreaky: bool=False,
    scales: tuple=[1,1],
    ax: plt.Axes=None,
    **kws
    ) -> plt.Axes:
    """Plot to show a table.

    Args:
        df1 (pd.DataFrame): input data.
        xlabel (str, optional): x label. Defaults to None.
        ylabel (str, optional): y label. Defaults to None.
        annot (bool, optional): show numbers. Defaults to True.
        cbar (bool, optional): show colorbar. Defaults to False.
        linecolor (str, optional): line color. Defaults to 'k'.
        linewidths (float, optional): line widths. Defaults to 1.
        cmap (str, optional): color map. Defaults to None.
        sorty (bool, optional): sort rows. Defaults to False.
        linebreaky (bool, optional): linebreak for y labels. Defaults to False.
        scales (tuple, optional): scale of the table. Defaults to [1,1].
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `sns.heatmap` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    print(df1.index.name,df1.columns.name)
    if xlabel is None and not df1.index.name is None:
        ylabel=df1.index.name
    if ylabel is None and not df1.columns.name is None:
        xlabel=df1.columns.name
#     print(xlabel,ylabel)
    from roux.viz.colors import make_cmap
    if sorty:
        df1=df1.loc[df1.sum(axis=1).sort_values(ascending=False).index,:]
    if linebreaky:
        df1.index=[linebreaker(s,break_pt=35) for s in df1.index]
    if ax is None:
        fig,ax=plt.subplots(figsize=[(df1.shape[1]*0.6)*scales[0],(df1.shape[0]*0.5)*scales[1]])
    ax=sns.heatmap(df1,
                cmap=make_cmap(['#ffffff','#ffffff']) if cmap is None else cmap,
                annot=annot,cbar=cbar,
                linecolor=linecolor,linewidths=linewidths,
                   ax=ax,
                   **kws,
               )
    ax.xaxis.set_label_position('top') 
    ax.xaxis.set_ticks_position('top')
    for tick in ax.yaxis.get_majorticklabels():
        tick.set_verticalalignment("center")
    ax.patch.set_edgecolor('k')
    ax.patch.set_linewidth(1) 
    set_ylabel(ax=ax,s=df1.index.name if ylabel is None else ylabel,xoff=0.05,yoff=0.01)
    return ax

def plot_crosstab(
    df1: pd.DataFrame,
    cols: list=None,
    alpha: float=0.05,
    method: str=None,#'chi2'|fe
    confusion: bool=False,
    rename_cols: bool=False,
    sort_cols: tuple=(True,True),    
    annot_pval: str='bottom',
    cmap: str='Reds',
    ax: plt.Axes=None,
    **kws,
    ) -> plt.Axes:
    """Plot crosstab table.

    Args:
        df1 (pd.DataFrame): input data
        cols (list, optional): columns. Defaults to None.
        alpha (float, optional): alpha for the stats. Defaults to 0.05.
        method (str, optional): method to check the association ['chi2','FE']. Defaults to None.
        rename_cols (bool, optional): rename the columns. Defaults to True.
        annot_pval (str, optional): annotate p-values. Defaults to 'bottom'.
        cmap (str, optional): colormap. Defaults to 'Reds'.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Raises:
        ValueError: `annot_pval` position should be the allowed one.

    Returns:
        plt.Axes: `plt.Axes` object.

    TODOs:
        1. Use `compare_classes` to get the stats.
    """
    if not cols is None:
        dplot=pd.crosstab(df1[cols[0]],df1[cols[1]])
    else:
        dplot=df1.copy()
    dplot=(dplot
        .sort_index(axis=0,ascending=sort_cols[0])
        .sort_index(axis=1,ascending=sort_cols[1])
          )        
    if dplot.shape!=(2,2) or method=='chi2':
        stat,pval,_,_=sc.stats.chi2_contingency(dplot)
        stat_label='${\chi}^2$'
    else:
        stat,pval=sc.stats.fisher_exact(dplot)
        stat_label='OR'
    if dplot.shape==(2,2) and rename_cols:
        dplot=dplot.rename(columns={True:dplot.columns.name,
                                   False:'not'},
                    index={True:dplot.index.name,
                          False:'not'},)
        dplot.columns.name=None
        dplot.index.name=None
        if 'not' in dplot.columns:
            dplot=dplot.loc[:,
                            [s for s in dplot.columns if s!='not']+['not']]
        if 'not' in dplot.index:
            dplot=dplot.loc[[s for s in dplot.index if s!='not']+['not'],
                            :]
    info(stat,pval)
    # dplot=dplot.sort_index(ascending=False,axis=1).sort_index(ascending=False,axis=0)
    ax=plot_table(dplot,
                    cmap=cmap,
                    ax=ax,
                    **kws,
                 )        
    if annot_pval:
        if annot_pval=='bottom':
            kws_set_label=dict(x=0.5,y=-0.2,ha='center',va='center',)
            linebreak=False
        elif annot_pval=='right':
            kws_set_label=dict(x=1,y=0,ha='left',va='bottom',)            
            linebreak=True
        else:
            raise ValueError(annot_pval)
        set_label(s=f"{stat_label}={stat:.1f}"+(', ' if not linebreak else '\n')+pval2annot(pval, alternative='two-sided', alpha=alpha, fmt='<', linebreak=False),
                 **kws_set_label,
                 ax=ax)
    if confusion:    
        ax=annot_confusion_matrix(dplot,ax=ax,
                                      off=0.5)
    return ax


from roux.global_imports import *

# set enrichment
def get_intersection_stats(df,coltest,colset,background_size=None):
    from roux.stat.binary import compare_bools_jaccard
    from scipy.stats import hypergeom,fisher_exact
    """
    :param background: size of the union (int)
    """
    hypergeom_p=hypergeom.sf(sum(df[coltest] & df[colset])-1,
                             len(df) if background_size is None else background_size,
                             df[colset].sum(),
                             df[coltest].sum(),)
    contigency=[[sum(df[coltest] & df[colset]),sum(df[coltest] & ~df[colset])],[sum(~df[coltest] & df[colset]),sum(~df[coltest] & ~df[colset])],]
    odds_ratio,fisher_exactp=fisher_exact(contigency,alternative='two-sided')
    jaccard=compare_bools_jaccard(df[coltest],df[colset])
    return hypergeom_p,fisher_exactp if jaccard!=0 else 1,odds_ratio,jaccard

def get_set_enrichment_stats(test,sets,background,fdr_correct=True):
    """
    test:
        get_set_enrichment_stats(background=range(120),
                        test=range(100),
                        sets={f"set {i}":list(np.unique(np.random.randint(low=100,size=i+1))) for i in range(100)})
        # background is int
        get_set_enrichment_stats(background=110,
                        test=unique(range(100)),
                        sets={f"set {i}":unique(np.random.randint(low=140,size=i+1)) for i in range(0,140,10)})                        
    """
    if isinstance(background,list):
        background_elements=np.unique(background)
        background_size=None
    elif isinstance(background,(int,float)):
        background_elements=list2union([test]+list(sets.values()))
        background_size=background
        if len(background_elements)>background_size:
            logging.error(f"invalid data type of background {type(background)}")
    else:
        logging.error(f"invalid data type of background {type(background)}")
    delement=pd.DataFrame(index=background_elements)
    delement.loc[np.unique(test),'test']=True
    for k in sets:
        delement.loc[np.unique(sets[k]),k]=True
    delement=delement.fillna(False)
    dmetric=pd.DataFrame({colset:get_intersection_stats(delement,'test',colset,background_size=background_size) for colset in delement if colset!='test'}).T.rename(columns=dict(zip([0,1,2,3],['hypergeom p-val','fisher_exact p-val','fisher_exact odds-ratio','jaccard index'])))
    if fdr_correct:
        from statsmodels.stats.multitest import multipletests
        for c in dmetric:
            if c.endswith(' p-val'):
                dmetric[f"{c} corrected"]=multipletests(dmetric[c], alpha=0.05, method='fdr_bh',
                                                        is_sorted=False,returnsorted=False)[1]
    return dmetric

def test_set_enrichment(tests_set2elements,test2_set2elements,background_size):
    from tqdm import tqdm
    from roux.lib.set import list2union
    dn2df={}
    for test1n in tqdm(tests_set2elements):
        for test2n in test2_set2elements:
            if len(tests_set2elements[test1n])!=0:
                dn2df[(test1n,test2n)]=get_set_enrichment_stats(test=tests_set2elements[test1n],
                                         sets={test2n:test2_set2elements[test2n]},
                                         background=background_size,
                                        fdr_correct=True,
                                        )
    denrich=pd.concat(dn2df,axis=0,names=['difference','test2 set'])
    return denrich

def get_paired_sets_stats(l1,l2):
    """
    overlap, intersection, union, ratio
    """
    if all([isinstance(l, list) for l  in [l1,l2]]):
        l=list(jaccard_index(l1,l2))
        l.append(get_ratio_sorted(len(l1),len(l2)))
        return l

def get_enrichment(df1,df2,
                   background,
           colid='gene id',
           colref='gene set id',
           colrefname='gene set name',
           colreftype='gene set type',
           colrank='rank',
           outd=None,
           name=None,
            # enrichr
           cutoff=0.05, # only used for plotting. 
            # prerank
           permutation_num=1000,
            # etc
            verbose=False,no_plot=True,
           **kws_prerank,
                  ):
    """
    
    :return leading edge gene ids: high rank first
    """
    
    import gseapy as gp
    name=get_name(df2,colreftype)
    
    import tempfile
    with tempfile.TemporaryDirectory() as p:
        outd=outd if not outd is None else p
        o1 = gp.enrichr(gene_list=df1[colid].unique().tolist(),
                         # or gene_list=glist
                         description=name,
                         gene_sets=df2.rd.to_dict([colref,colid]),
                         background=background, # or the number of genes, e.g 20000
                         outdir=f'{outd}/{name}',
                         cutoff=cutoff, # only used for plotting.
                         verbose=verbose,
                         no_plot=no_plot,
    #                      **kws,
                         )
        df3=o1.results
    df3=df3.rename(columns={'Term':colref,
                             'P-value':'P (FE test)',
                             'Adjusted P-value':'P (FE test, FDR corrected)',
                             'Genes':f'{colid}s'},errors='raise')
    df_=df3['Overlap'].str.split('/',expand=True).rename(columns={0:f"{colid}s overlap",
                                                         1:f"{colid}s per {colref}"}).applymap(int)
    df3=df3.join(df_)
#     df3['overlap %']=df3['Overlap'].apply(eval)*100    
    df3['overlap %']=df3.apply(lambda x: (x[f"{colid}s overlap"]/x[f"{colid}s per {colref}"])*100,axis=1)
    df3=df3.drop(['Gene_set','Overlap',],axis=1)
    info('enrichr: '+perc_label(sum(df3['P (FE test, FDR corrected)']<=0.05),len(df3)))
    if not colrank in df1:
        if colrefname in df2:
            df3[colrefname]=df3[colref].map(df2.set_index(colref)[colrefname].drop_duplicates())
        return df3
    if df3[f"{colid}s overlap"].max()<2:
        logging.error("df3[f'{colid}s overlap'].max()<2 # can not run prerank")
        return df3
    with tempfile.TemporaryDirectory() as p:
        outd=outd if not outd is None else p    
        o2 = gp.prerank(rnk=df1.loc[:,[colid,colrank]],
                         gene_sets=df2.rd.to_dict([colref,colid]),
                         min_size=2,
                         max_size=df3[f"{colid}s overlap"].max(),
                         processes=1,
                         permutation_num=permutation_num, # reduce number to speed up testing
                         ascending=False, # look for high number 
                         outdir=f'{outd}/{name}',
                         pheno_pos='high',
                         pheno_neg='low',
                         format='png',
                        no_plot=no_plot,
                        seed=1,
                        graph_num=1,
                        **kws_prerank,
                       )
        df4=o2.res2d.reset_index()
#     es	nes	pval	fdr	geneset_size	matched_size	genes	ledge_genes
    df4=df4.rename(columns={'Term':colref,
                            'es':'enrichment score',
                            'nes':'normalized enrichment score', 
                             'pval':'P (GSEA test)',
                            'fdr':'FDR (GSEA test)',
                            'matched_size':f"{colid}s overlap",
                            'geneset_size':f"{colid}s per {colref}",
                             'ledge_genes':f'{colid}s leading edge',
                           },
                   errors='raise')
    df4['overlap %']=df4.apply(lambda x: (x[f"{colid}s overlap"]/x[f"{colid}s per {colref}"])*100,axis=1)
    df4=df4.drop(['genes',],axis=1)
    info('preraked: '+perc_label(sum(df4['P (GSEA test)']<=0.05),len(df4)))
    
    df5=df3.merge(df4,
              on='gene set id',
              how='left',
             validate="1:1",
                 suffixes=['',' (prerank)'])
    df5=df5.drop([
#                 f'{colid}s',
                  f'{colid}s per {colref} (prerank)',f'{colid}s overlap (prerank)','overlap % (prerank)'],axis=1)
    return df5

def get_enrichments(df1,
                    df2,
                    background,
                    coltest='subset',
                    colid='gene id',
                    colref='gene set id',
                    colreftype='gene set type',
                    fast=False,
                    **kws):
    """
    :param df1: test sets
    :param df2: reference sets
    """
    return getattr(df1.groupby(coltest),'progress_apply' if not fast else 'parallel_apply')(lambda df_: df2.groupby(colreftype).apply(lambda df: get_enrichment(df_,df,background=background,**kws)).reset_index(0)).reset_index(0)
    


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from roux.viz.ax_ import *

def plot_barh(
    df1: pd.DataFrame,
    colx: str,
    coly: str,
    colannnotside: str=None,
    x1: float=None,
    offx: float=0,
    ax: plt.Axes=None,
    **kws
    ) -> plt.Axes:
    """Plot horizontal bar plot with text on them.

    Args:
        df1 (pd.DataFrame): input data.
        colx (str): x column.
        coly (str): y column.
        colannnotside (str): column with annotations to show on the right side of the plot.
        x1 (float): x position of the text.
        offx (float): x-offset of x1, multiplier.
        color (str): color of the bars.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `barh` function.

    Returns:
        plt.Axes: `plt.Axes` object.
    """    
    df1['y']=range(len(df1[coly]))
    if ax is None:
        fig,ax=plt.subplots(figsize=[4.45,len(df1)*0.33])
    if x1 is None:
        x1=get_axlims(ax)['x']['min']+(get_axlims(ax)['x']['len']*(0.05*(1-offx)))
    ax=df1.set_index(coly)[colx].plot.barh(width=0.8,ax=ax,**kws)
    df1['yticklabel']=[t.get_text() for t in ax.get_yticklabels()]
    _=df1.apply(lambda x: ax.text(x1,x['y'],x['yticklabel'],va='center'),axis=1)
    if colannnotside is None:
        colannnotside=colx
    _=df1.apply(lambda x: ax.text(ax.get_xlim()[1],x['y'],f" {x[colannnotside]}",va='center'),axis=1)
    ax.get_yaxis().set_visible(False)
    ax.invert_yaxis()
    ax.set(xlabel=colx,)
    return ax

def plot_value_counts(
    df: pd.DataFrame,
    col: str,
    logx: bool=False,
    kws_hist: dict={'bins':10},
    kws_bar: dict={},
    grid: bool=False,
    axes: list=None,
    fig: object=None,
    hist: bool=True
    ):
    """Plot pandas's `value_counts`. 

    Args:
        df (pd.DataFrame): input data `value_counts`.
        col (str): column with counts.
        logx (bool, optional): x-axis on log-scale. Defaults to False.
        kws_hist (_type_, optional): parameters provided to the `hist` function. Defaults to {'bins':10}.
        kws_bar (dict, optional): parameters provided to the `bar` function. Defaults to {}.
        grid (bool, optional): show grids or not. Defaults to False.
        axes (list, optional): list of `plt.axes`. Defaults to None.
        fig (object, optional): figure object. Defaults to None.
        hist (bool, optional): show histgram. Defaults to True.
    """
    dplot=pd.DataFrame(df[col].value_counts()).sort_values(by=col,ascending=True)
    figsize=(4, len(dplot)*0.4)
    if axes is None:
        if hist:
            fig, axes = plt.subplots(2,1, sharex=True, 
                                       gridspec_kw={'height_ratios': [1, 6]},
                                       figsize=figsize,
                                    )
        else:
            fig, axes = plt.subplots(1,1, 
                                    figsize=figsize,
                                    )
    if hist:
        _=dplot.plot.hist(ax=axes[0],legend=False,**kws_hist
        #                   orientation="horizontal"
                         )
        axbar=axes[1]
    else:
        axbar=axes
    _=dplot.plot.barh(ax=axbar,legend=False,**kws_bar)
    axbar.set_xlabel('count')
    from roux.lib.str import linebreaker
    axbar.set_ylabel(col.replace(' ','\n'))
    if logx:
        if hist:
            for ax in axes.flat:
                ax.set_xscale("log")
                if grid:
                    ax.set_axisbelow(False)
        else:
            axes.set_xscale("log")
            if grid:
                axes.set_axisbelow(False)

def plot_barh_stacked_percentage(
    df1: pd.DataFrame,
    coly: str,
    colannot: str,
    color: str=None,
    yoff: float=0,
    ax: plt.Axes = None,
    ) -> plt.Axes:
    """Plot horizontal stacked bar plot with percentages.

    Args:
        df1 (pd.DataFrame): input data. values in rows sum to 100%.
        coly (str): y column. yticklabels, e.g. retained and dropped.
        colannot (str): column with annotations.
        color (str, optional): color. Defaults to None.
        yoff (float, optional): y-offset. Defaults to 0.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """

    from roux.viz.ax_ import get_ticklabel2position
    from roux.viz.colors import get_colors_default
    if color is None:
        color=get_colors_default()[0]
    ax=plt.subplot() if ax is None else ax
    df2=df1.set_index(coly).apply(lambda x: (x/sum(x))*100, axis=1)
    ax=df2.plot.barh(stacked=True,ax=ax)
    ticklabel2position=get_ticklabel2position(ax,'y')
    from roux.viz.colors import saturate_color
    _=df2.reset_index().apply(lambda x: ax.text(1,
                                                  ticklabel2position[x[coly]]-yoff,
                                                  f"{x[colannot]:.1f}%",ha='left',va='center',
                                                 color=saturate_color(color,2),
                                               ),
                                axis=1)
    ax.legend(bbox_to_anchor=[1,1],title=df1.columns.name)
    d1=df1.set_index(coly).T.sum().to_dict()
    ax.set(xlim=[0,100],xlabel='%',
          yticklabels=[f"{t.get_text()}\n(n={d1[t.get_text()]})" for t in ax.get_yticklabels()])
    return ax

def plot_bar_serial(
    d1: dict,
    polygon: bool=False,
    polygon_x2i: float=0,
    labelis: list=[],
    y: float=0,
    ylabel: str=None,
    off_arrowy: float=0.15,
    kws_rectangle=dict(height=0.5,linewidth=1),
    ax: plt.Axes = None,
    ) -> plt.Axes:
    """Barplots with serial increase in resolution.

    Args:
        d1 (dict): dictionary with the data.
        polygon (bool, optional): show polygon. Defaults to False.
        polygon_x2i (float, optional): connect polygon to this subset. Defaults to 0.
        labelis (list, optional): label these subsets. Defaults to [].
        y (float, optional): y position. Defaults to 0.
        ylabel (str, optional): y label. Defaults to None.
        off_arrowy (float, optional): offset for the arrow. Defaults to 0.15.
        kws_rectangle (_type_, optional): parameters provided to the `rectangle` function. Defaults to dict(height=0.5,linewidth=1).
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """

    kws1=dict(
    xs=[(list(d1.values())[i]/sum(list(d1.values())))*100 for i in range(len(d1))],
    cs=['#f55f5f', '#D3DDDC'] if len(d1)==2 else ['#f55f5f','#e49e9d', '#D3DDDC'],
    labels=[],
    size=sum(list(d1.values())),
    xmax=100,
    y=y,
    ax=None,
    )
    
    import matplotlib.patches as patches
    l1=[
        patches.Rectangle((0,kws1['y']), kws1['xmax'],  
                          fc='none', ec='k',alpha=1,zorder=2,**kws_rectangle),
    ]
    x=0
    for i,(x_,c,s) in enumerate(zip(kws1['xs'],kws1['cs'],d1.keys())):
        print(x,x_)
        l1.append(patches.Rectangle((x, kws1['y']), x_, 
                          fc=c, ec='none',alpha=1,**kws_rectangle))
        if i in labelis:
            ax.text(x if i!=len(d1.keys())-1 else x+x_,
                    kws1['y']+kws_rectangle['height']*0.5,
                    f"{s} ({x_:.0f})%",
                    ha='left' if i!=len(d1.keys())-1 else 'right',
                    va='center')
        x+=x_
        if polygon:
            if polygon_x2i==i:
                l1.append(patches.Polygon(xy=[(0,kws1['y']),
                                              (x,kws1['y']),
                                              (100,kws1['y']-1),
                                              (0,kws1['y']-1),], closed=True,fc=[0.95,0.95,0.95],
                                          ec='gray',
                                          lw=0.1,
                                          zorder=-2))        
    #     break
    s=num2str(kws1['size'],magnitude=True,decimals=1)
    ax.annotate(s=s, xy=(0,kws1['y']-off_arrowy), xytext=(50,kws1['y']-off_arrowy), arrowprops=dict(arrowstyle='->',shrinkA=0,shrinkB=0,color='k'),zorder=-1,
               va='center',ha='center')
    ax.annotate(s=' '*(len(s)*2), xy=(100,kws1['y']-off_arrowy), xytext=(50,kws1['y']-off_arrowy), arrowprops=dict(arrowstyle='->',shrinkA=0,shrinkB=0,color='k'),zorder=-1,
               va='center',ha='center')
    _=[ax.add_patch(o) for o in l1]
    if not ylabel is None:
        ax.text(-2.5,kws1['y']+kws_rectangle['height']*0.5,ylabel,ha='right',va='center')
            
    return ax

def plot_barh_stacked_percentage_intersections(
    df0: pd.DataFrame,
    colxbool: str,
    colybool: str,
    colvalue: str,
    colid: str,
    colalt: str,
    colgroupby: str,
    coffgroup: float=0.95,
    ax: plt.Axes = None,
    ) -> plt.Axes:
    """Plot horizontal stacked bar plot with percentages and intesections.

    Args:
        df0 (pd.DataFrame): input data.
        colxbool (str): x column.
        colybool (str): y column.
        colvalue (str): column with the values.
        colid (str): column with ids.
        colalt (str): column with the alternative subset.
        colgroupby (str): column with groups.
        coffgroup (float, optional): cut-off between the groups. Defaults to 0.95.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.

    Examples:
        Parameters:
            colxbool='paralog',
            colybool='essential',
            colvalue='value',
            colid='gene id',
            colalt='singleton',
            coffgroup=0.95,
            colgroupby='tissue',

    """

    ##1 threshold for value by group
    def apply_(df):
        coff=np.quantile(df.loc[df[colybool],colvalue],coffgroup)
        df[colybool]=df[colvalue]<coff
        return df
    df1=df0.groupby(colgroupby).progress_apply(apply_)
    ##2 % 
    df2=df1.groupby([colid,colxbool]).agg({colybool: perc}).reset_index().rename(columns={colybool:f'% {colgroupby}s with {colybool}'},
                                                                                        errors='raise')
    coly=f"% of {colgroupby}s"
    ##3 bin y
    df2[coly]=pd.cut(df2[f'% {colgroupby}s with {colybool}'],bins=pd.interval_range(0,100,4),)
    ##3 % sum
    df3=df2.groupby(coly)[colxbool].agg([sum]).rename(columns={'sum':colxbool})
    dplot=df3.join(df2.groupby(coly).size().to_frame('total'))
    dplot[colalt]=dplot['total']-dplot[colxbool]
    dplot.index=[str(i) for i in dplot.index]
    dplot.index.name=coly
    dplot.columns.name=f"{colid} type"
    dplot=dplot.sort_values(coly,ascending=False)
    dplot=dplot.reset_index()
#     from roux.viz.bar import plot_barh_stacked_percentage
    fig,ax=plt.subplots(figsize=[3,3])
    plot_barh_stacked_percentage(df1=dplot.loc[:,[coly,colxbool,colalt]],
                                coly=coly,
                                colannot=colxbool,
                                ax=ax)
    set_ylabel(ax)
    ax.set(xlabel=f'% of {colid}s',
          ylabel=None)
    return ax

# redirections, to be deprecated in the future
from roux.viz.sets import plot_intersections

## plotly
def to_input_data_sankey(df0,
                         colid,
                         cols_groupby=None,
                        colall='all',
                         remove_all=False,
                        ):
    """
    """
    if cols_groupby is None:
        cols_groupby=list(set(df0.columns.tolist())-set([colid]))
    df0=(df0
        .log.dropna(subset=cols_groupby,
                       how='all',
                   )
        )
    if not colall in cols_groupby:
        df0=df0.assign(**{colall:colall})
        cols_groupby=[colall]+cols_groupby
    for col in cols_groupby:
        if df0[col].dtype=='bool':
            df0[col]=df0[col].map({True:col,False:np.nan}) 
    ## map labels    
    # print(cols_groupby)
    # print(np.lib.stride_tricks.sliding_window_view(cols_groupby[1:],2))
    df1=pd.concat([
        df0.groupby(list(cols))[colid].nunique() for cols in np.lib.stride_tricks.sliding_window_view(cols_groupby[1:],2) # if the counts do not reduce serially: itertools.combinations(cols_groupby[1:],2)
    ],
    axis=0
    ).to_frame('count').rename_axis(['source','target'],axis=0).reset_index().drop_duplicates()
    # df1

    # to set the correct number of count for the 1st subset bar in case where the counts decrease
    df2=pd.concat([
        df0.groupby(list(cols))[colid].nunique() for cols in itertools.product([colall],[cols_groupby[1]])
    ],
    axis=0
    ).to_frame('total').rename_axis(['source','target'],axis=0).reset_index().drop_duplicates()
    # df2

    df3=(
    df2
    .merge(df1.groupby('target')['count'].sum().to_frame('substract').reset_index(),
              on=['target'],
             how='left',
             )
    .assign(**{
        'substract':lambda x: x['substract'].fillna(0),    
        'count':lambda x: x['total']-x['substract']})
    .append(df1)
    )
    if remove_all:
        # to set the correct number of count for the 1st subset bar in case where the counts decrease
        assert len(set(df0.loc[:,cols_groupby].replace(False,np.nan).apply(lambda x: x.notnull().sum())[1:3].tolist()))==1
        df3=df3.loc[(df3['source']!=colall) & (df3['target']!=colall),:]
    # map ids
    labels=list(pd.unique(df3['source'].unique().tolist()+df3['target'].unique().tolist()))
    d_={s:i for i,s in enumerate(labels)}
    for c in ['source','target']:
        df3[c+' id']=df3[c].map(d_)
    return df3#,labels

def plot_sankey(
    df1,
    cols_groupby=None,
    hues=None, # dict
    node_color=None,
    link_color=None,
    info=None,
    x=None,
    y=None,
    colors=None,
    hovertemplate=None,
    text_width=20,
    convert=True,
    width=400,
    height=400,
    outp=None,
    validate=True,
    test=False,
    **kws,
    ):
    if convert:
        df2=to_input_data_sankey(df1,
                                 cols_groupby=cols_groupby,
                                 **kws)
        id2n_=df1.apply(lambda x: len(df1)-x.isnull().sum())
    id2n=df2.groupby(['target'])['count'].sum().astype(int)#.to_dict()
    # print(id2n)
    # print(id2n_)
    if convert and validate:    
        assert all(id2n_.loc[id2n.index]==id2n), 'sizes of the sets changed after `to_input_data_sankey`?'
    labels=list(pd.unique(df2['source'].unique().tolist()+df2['target'].unique().tolist()))
    if not hues is None:
        from roux.viz.colors import get_val2color,to_hex
        val2color,legend2color=get_val2color(pd.Series(hues))
        # df2[f"{col} color"]=df2[col].map(hues)
        colors=[to_hex(val2color[hues[k]]) if k in hues else "#888888" for k in labels]
    if not info is None:
        customdata=[info[k] if k in info else k for k in labels]
    else:
        customdata=None
    from roux.lib.str import linebreaker
    labels=[linebreaker(x +(f" ({id2n[x]})" if x in id2n else ''),text_width,sep='<br>') for x in labels]
    ## get x and y values    
    import plotly.io as pio
    pio.renderers.default = 'iframe'
    import plotly.graph_objects as go
    fig = go.Figure(data=[go.Sankey(
        arrangement='freeform',
        # arrangement='fixed',
        node = dict(
            pad = 0,
            thickness = 7,
            line = dict(
                color = '#FFFFFF',
                width = 0.1,
                ),
            x=x,
            y=y,
            # x=np.repeat(1,len(labels)),
            # y=np.repeat(1,len(labels)),
            label =labels,
            color=colors,
            customdata=customdata,
            hovertemplate=hovertemplate,#'%{customdata} value %{value}<extra></extra>',
            ),
        link = dict(
            source = df2['source id'],
            target = df2['target id'],
            value = df2['count'],
            # color=link_color,
            ),
      )])
    # fig.for_each_annotation(lambda a: a.update(x=a.x-1))
    # fig.for_each_annotation(lambda a: a.update(text=a.text.split("n")[-1]))
    # fig.update_annotations(align='left')
    # #x axis
    # fig.update_xaxes(visible=True)
    # #y axis    
    # fig.update_yaxes(visible=True)    
    fig.update_layout(
                  font_size=10,
                  autosize=False,
                  width=width,
                  height=height,
                    # xaxis= {
                    #     'range': [0.2, 1],
                    #     'showgrid': True, # thin lines in the background
                    #     'zeroline': False, # thick line at x=0
                    #     'visible': True,  # numbers below
                    # }, # the same for yaxis
                 )
    
    if not outp is None:
        fig.write_image(outp,format=Path(outp).suffix[1:], engine="kaleido")
    else:
        fig.show()
    return fig


from roux.lib.dfs import *
import matplotlib.pyplot as plt
import scipy as sc

# scikit learn       
def check_clusters(df: pd.DataFrame):
    """Check clusters.

    Args:
        df (DataFrame): dataframe.

    """
    return (df.groupby(['cluster #']).agg({'silhouette value':np.max})['silhouette value']>=df['silhouette value'].mean()).all()

def get_clusters(X: np.array,n_clusters: int,random_state=88,
                 params={},
                 test=False) -> dict:
    """Get clusters.

    Args:
        X (np.array): vector
        n_clusters (int): int
        random_state (int, optional): random state. Defaults to 88.
        params (dict, optional): parameters for the `MiniBatchKMeans` function. Defaults to {}.
        test (bool, optional): test. Defaults to False.

    Returns:
        dict: 
    """
    from sklearn import cluster,metrics
    kmeans = cluster.MiniBatchKMeans(n_clusters=n_clusters,
                             random_state=random_state,
                             **params,
                            ).fit(X)
    clusters=kmeans.predict(X)
    ds=pd.Series(dict(zip(X.index,clusters)))
    ds.name='cluster #'
    df=pd.DataFrame(ds)
    df.index.name=X.index.name
    df['cluster #'].value_counts()
    # Compute the silhouette scores for each sample
    df['silhouette value'] = metrics.silhouette_samples(X, clusters)
#     if test:
#         print(f"{n_clusters} cluster : silhouette average score {df['silhouette value'].mean():1.2f}, ok?: {is_optimum}, random state {random_state}")
    if not check_clusters:
        logging.warning(f"{n_clusters} cluster : silhouette average score {df['silhouette value'].mean():1.2f}, ok?: {is_optimum}, random state {random_state}")
    dn2df={'clusters':df.reset_index(),
           'inertia':kmeans.inertia_,
           'centers':pd.DataFrame(kmeans.cluster_centers_,index=range(n_clusters),columns=X.columns).rename_axis(index='cluster #')#.stack().reset_index().rename(columns={'level_1':'variable',0:'value'}),
          }
    return dn2df

def get_n_clusters_optimum(df5: pd.DataFrame,test=False) -> int:
    """Get n clusters optimum.

    Args:
        df5 (DataFrame): input dataframe
        test (bool, optional): test. Defaults to False.

    Returns:
        int: knee point.
    """
    from kneed import KneeLocator
    kn = KneeLocator(x=df5['total clusters'], y=df5['inertia'], curve='convex', direction='decreasing')
    if test:
        import matplotlib.pyplot as plt
        kn.plot_knee()
        plt.title(f"knee point={kn.knee}")
    return kn.knee
        
def plot_silhouette(df: pd.DataFrame,n_clusters_optimum=None,ax=None):
    """Plot silhouette

    Args:
        df (DataFrame): input dataframe.
        n_clusters_optimum (int, optional): number of clusters. Defaults to None:int.
        ax (axes, optional): axes object. Defaults to None:axes.

    Returns:
        ax (axes, optional): axes object. Defaults to None:axes.
    """
    import matplotlib.pyplot as plt
    import seaborn as sns
    ax=plt.subplot() if ax is None else ax
    ax=sns.violinplot(data=df.groupby(['total clusters','cluster #']).agg({'silhouette value':np.mean}).reset_index(),
                     y='silhouette value',x='total clusters',
                     color='salmon',alpha=0.7,
                     ax=ax)
    ax=sns.pointplot(data=df.groupby('total clusters').agg({'silhouette value':np.mean}).reset_index(),
                     y='silhouette value',x='total clusters',
                     color='k',
                     ax=ax)
    if not n_clusters_optimum is None:
        ax.annotate('optimum', 
                xy=(n_clusters_optimum-int(ax.get_xticklabels()[0].get_text()),
                    ax.get_ylim()[0]),  
                xycoords='data',
                    xytext=(+50, +50), textcoords='offset points',
                    arrowprops=dict(arrowstyle="->",ec='k',
                                    connectionstyle="angle3,angleA=0,angleB=-90"),
                    )
    ax.set_xlabel('clusters')  
    return ax

def get_clusters_optimum(X: np.array,n_clusters=range(2,11),
                         params_clustering={},
                         test=False,
                        ) -> dict:
    """Get optimum clusters.

    Args:
        X (np.array): samples to cluster in indexed format.
        n_clusters (int, optional): _description_. Defaults to range(2,11).
        params_clustering (dict, optional): parameters provided to `get_clusters`. Defaults to {}.
        test (bool, optional): test. Defaults to False.

    Returns:
        dict: _description_
    """
    dn2d={}
    for n in n_clusters:
        dn2d[n]=get_clusters(X=X,n_clusters=n,test=test,params=params_clustering)
    df1=pd.DataFrame(pd.Series({k:dn2d[k]['inertia'] for k in dn2d}),columns=['inertia']).rename_axis(index='total clusters').reset_index()
    # TODO identify saturation point in the intertia plot for n_clusters_optimum
    n_clusters_optimum=get_n_clusters_optimum(df1,test=test)
    #
    dn2df={dn:pd.concat({k:dn2d[k][dn] for k in dn2d},axis=0,names=['total clusters']).reset_index() for dn in ['clusters','centers']}
    if not check_clusters(dn2df['clusters']):
        logging.warning('low silhoutte scores')
        return
    if test:
        import matplotlib.pyplot as plt
        plt.figure()
        plot_silhouette(df=dn2df['clusters'],n_clusters_optimum=None)
    # make output
    dn2df={dn:dn2df[dn].loc[(dn2df[dn]['total clusters']==n_clusters_optimum),:].drop(['total clusters'],axis=1) for dn in dn2df}
    return dn2df

def get_gmm_params(g,x,
                         n_clusters=2,
                         test=False,
                        ):
    """Intersection point of the two peak Gaussian mixture Models (GMMs).
    
    Args:
        out (str): `coff` only or `params` for all the parameters.
        
    """
    assert n_clusters==2
    weights = g.weights_
    means = g.means_
    covars = g.covariances_
    stds=np.sqrt(covars).ravel().reshape(n_clusters,1)
    # logging.info(f'weights {weights}')
    f = x.reshape(-1,1)
    x.sort()
    two_pdfs = sc.stats.norm.pdf(np.array([x,x]), means, stds)
    mix_pdf = np.matmul(weights.reshape(1,n_clusters), two_pdfs)
    return mix_pdf,two_pdfs,means,weights
    
def get_gmm_intersection(x,two_pdfs,means,weights,test=False):
    from roux.stat.solve import get_intersection_locations
    idxs=get_intersection_locations(y1=two_pdfs[0]*weights[0],
                                    y2=two_pdfs[1]*weights[1],
                                    test=False,x=x)
    x_intersections=x[idxs]
    if test: logging.info(f'intersections {x_intersections}')
    ms=sorted([means[0][0],means[1][0]])
    if len(x_intersections)>1:
        if test:
            logging.info(x_intersections)
            logging.info(ms)
            logging.info([i for i in x_intersections if i>ms[0] and i<ms[1]])
        coffs_=[i for i in x_intersections if i>ms[0] and i<ms[1]]
        if len(coffs_)!=0:
            coff=coffs_[0]
        else:
            coff=None
        if test:
            logging.info(coff)
    else:
        coff=x_intersections[0]
    return coff

def cluster_1d(ds: pd.Series,
                n_clusters: int,
                clf_type='gmm',
                random_state=1,
                test=False,
                returns=['coff'],
                **kws_clf) -> dict:
    """Cluster 1D data.

    Args:
        ds (Series): series.
        n_clusters (int): number of clusters.
        clf_type (str, optional): type of classification. Defaults to 'gmm'.
        random_state (int, optional): random state. Defaults to 88.
        test (bool, optional): test. Defaults to False.
        returns (list, optional): return format. Defaults to ['df','coff','ax','model'].
        ax (axes, optional): axes object. Defaults to None.

    Raises:
        ValueError: clf_type

    Returns:
        dict: _description_
    """
    assert not ds._is_view, "input series should be a copy not a view"
    x=ds.to_numpy()
    X=x.reshape(-1,1)
    if clf_type.lower()=='gmm':
        from sklearn.mixture import GaussianMixture
        model = GaussianMixture(n_components=n_clusters,random_state=random_state)
    elif clf_type.lower()=='kmeans':
        from sklearn.cluster import KMeans
        model=KMeans(n_clusters=n_clusters,**kws_clf).fit(X,)
    else:
        raise ValueError(clf_type)
    ## fit and predic
    labels =model.fit_predict(X)
    assert model.converged_
    df=pd.DataFrame({'value':x,
                     'label':labels==1})
    if clf_type=='gmm':
        mix_pdf,two_pdfs,means,weights=get_gmm_params(g=model,x=x,
                             n_clusters=n_clusters,
                             test=test,
                            )
        coff=get_gmm_intersection(x,two_pdfs,means,weights,test=test)
    d={}
    for k in returns:
        d[k]=locals()[k]        
    if test:
        if clf_type=='gmm':
            from roux.viz.dist import plot_gmm
            ax=plot_gmm(x,coff,mix_pdf,two_pdfs,weights,n_clusters=n_clusters,)
        else:
            coffs=df.groupby('label')['value'].agg(min).values
            for c in coffs:
                if  c > df['value'].quantile(0.05) and c < df['value'].quantile(0.95):
                    coff=c
                    break
            logging.info(f"coff:{c}; selected from {coffs}")
            ax.axvline(coff,color='k')
            ax.text(coff,ax.get_ylim()[1],f"{coff:.1f}",ha='center',va='bottom')            
    return d

## umap
def get_pos_umap(df1,spread=100,
                 test=False,k='',
                 **kws):
    """Get positions of the umap points.

    Args:
        df1 (DataFrame): input dataframe
        spread (int, optional): spead extent. Defaults to 100.
        test (bool, optional): test. Defaults to False.
        k (str, optional): number of clusters. Defaults to ''.

    Returns:
        DataFrame: output dataframe.
    """
    import umap
    reducer = umap.UMAP(spread=spread,*kws)
    embedding = reducer.fit_transform(df1)
    if test:
        plt.scatter(
            embedding[:, 0],
            embedding[:, 1],
            c=['r' if k in s else 'k' for s in df1.index.get_level_values(0)],
        )
        plt.gca().set_aspect('equal', 'datalim')
    df2=pd.DataFrame(embedding,
                    columns=['x','y'])
    df2.index=df1.index
    return df2.reset_index()


"""For processing individual pandas DataFrames/Series"""
import pandas as pd
import numpy as np
import logging
from icecream import ic
from roux.lib import to_rd

@to_rd
def get_name(df1,cols=None,coff=2,out=None):
    """Gets the name of the dataframe. 
    
    Especially useful within `groupby`+`pandarellel` context.

    Parameters:
        df1 (DataFrame): input dataframe.
        cols (list): list groupby columns. 
        coff (int): cutoff of unique values to infer the name.
        out (str): format of the output (list|not).

    Returns:
        name (tuple|str|list): name of the dataframe. 
    """    
    if hasattr(df1,'name') and cols is None:
        name=df1.name
        name=name if isinstance(name,str) else list(name)
    elif not cols is None:
        name=df1.iloc[0,:][cols]
    else:
        l1=get_constants(df1.select_dtypes(object))
        if len(l1)==1:
            from roux.lib.set import list2str
            name=list2str(df1[l1[0]].unique())
        elif len(l1)<=coff:
            name=sorted(l1)
        elif len(l1)==0:
            return
        else:
            logging.warning(f"possible names in here?: {','.join(l1)}")
            return
    if out=='list':
        if isinstance(name,str):
            name=[name]
    return name
            
@to_rd            
def get_groupby_columns(df_): 
    """Get the columns supplied to `groupby`.

    Parameters:
        df_ (DataFrame): input dataframe.

    Returns:
        columns (list): list of columns.
    """
    return df_.apply(lambda x: all(x==df_.name)).loc[lambda x: x].index.tolist()

@to_rd
def get_constants(df1):
    """Get the columns with a single unique value.

    Parameters:
        df1 (DataFrame): input dataframe.

    Returns:
        columns (list): list of columns.
    """
    return df1.nunique().loc[lambda x: x==1].index.tolist()

## delete unneeded columns
@to_rd
def drop_unnamedcol(df):
    """Deletes the columns with "Unnamed" prefix.

    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.
    """
    cols_del=[c for c in df.columns if 'Unnamed' in c]
    return df.drop(cols_del,axis=1)
### alias
delunnamedcol=drop_unnamedcol

@to_rd
def drop_levelcol(df):
    """Deletes the potentially temporary columns names with "level" prefix.

    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.
    """
    cols_del=[c for c in df.columns if 'level' in c]
    return df.drop(cols_del,axis=1)

@to_rd
def drop_constants(df):
    """Deletes columns with a single unique value.

    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.
    """    
    cols_del=get_constants(df)
    logging.warning(f"dropped columns: {', '.join(cols_del)}")
    return df.drop(cols_del,axis=1)

@to_rd
def dropby_patterns(df1,l1=None,test=False):
    """Deletes columns containing substrings i.e. patterns.

    Parameters:
        df1 (DataFrame): input dataframe.
        l1 (list): list of substrings.
        test (bool): verbose. 
        
    Returns:
        df1 (DataFrame): output dataframe.
    """
    if l1 is None or l1==[]:
        return df1
    if isinstance(l1,str):
        l1=[l1]
    s0='|'.join(l1).replace('(','\(').replace(')','\)')
    s1=f"^.*({s0}).*$"
    cols=df1.filter(regex=s1).columns.tolist()
    if test: info(s1)
    assert(len(cols)!=0)
    logging.info('columns dropped:'+','.join(cols))
    return df1.log.drop(labels=cols,axis=1)

@to_rd
def drop_inflates(df1,
                    col,
                    cols_index,
                  test=False,
                   ):
    """Deletes columns with high number of duplicates.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        col (str): column with values.
        cols_index (list): index columns.
        test (bool): verbose. 
        
    Returns:
        df1 (DataFrame): output dataframe.
    
    Notes:
        Under development.
    """
    logging.warning("UNDER DEVELOPMENT")
    df_=df1.rd.check_duplicated(cols_index,out='df')
    ## how many times duplicated?
    ds1_=df_[col].value_counts()
    ## unique number of times duplicated? e.g. x times whereas others are consistently y times
    ds2_=ds1_.value_counts()
    if test: 
        print(ds1_.head())
        print(ds2_)
    if len(ds2_)==1:
        logging.warning("skipping, inflates are consistent")
        logging.info(ds2_)
        return df1
    if len(ds2_.value_counts())==1:
        logging.warning("skipping, inflates are consistent")
        logging.info(ds2_)
        return df1        
    ## drop the unique number of times duplicated
    dropped_items=ds1_.loc[ds1_.isin(ds2_.loc[lambda x: x==1].index)].index.tolist()
    logging.info(f"dropped_items: {', '.join(dropped_items)}")
    df1=df1.loc[~(df1[col].isin(dropped_items)),:]
    assert(all(df1.rd.check_inflation([col])<1))
    return df1

@to_rd
def flatten_columns(df,**kws):
    """Multi-index columns to single-level.

    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.

    Keyword Arguments:
        kws (dict): parameters provided to `coltuples2str` function.    
    """    
    df.columns=coltuples2str(df.columns,**kws)
    return df

@to_rd
def lower_columns(df):
    """Column names of the dataframe to lower-case letters.

    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.
    """    
    df.columns=df.columns.str.lower()
    return df

@to_rd
def renameby_replace(df,replaces,ignore=True,**kws):
    """Rename columns by replacing sub-strings.

    Parameters:
        df (DataFrame): input dataframe.
        replaces (dict|list): from->to format or list containing substrings to remove.
        ignore (bool): if True, not validate the successful replacements.
        
    Returns:
        df (DataFrame): output dataframe.

    Keyword Arguments:
        kws (dict): parameters provided to `replacemany` function.    
    """    
    from roux.lib.str import replacemany
    df.columns=[replacemany(c,replaces,ignore=ignore,**kws) for c in df]
    return df


@to_rd
def clean_columns(df):
    """Standardise columns.

    Steps:
        1. Strip flanking white-spaces.
        2. Lower-case letters.
    
    Parameters:
        df (DataFrame): input dataframe.
        
    Returns:
        df (DataFrame): output dataframe.
    """
    df.columns=df.columns.str.strip().str.rstrip().str.lower()
    return df

@to_rd
def clean(df,cols=[],
          drop_constants=False,
          drop_unnamed=True,
          verb=False,
         ):
    """Deletes potentially temporary columns.

    Steps:
        1. Strip flanking white-spaces.
        2. Lower-case letters.
    
    Parameters:
        df (DataFrame): input dataframe.
        drop_constants (bool): whether to delete the columns with a single unique value. 
        drop_unnamed (bool): whether to delete the columns with 'Unnamed' prefix. 
        verb (bool): verbose.
        
    Returns:
        df (DataFrame): output dataframe.
    """
    cols_del=df.filter(regex="^(?:index|level|Unnamed|chunk|_).*$").columns.tolist()+df.filter(regex="^.*(?:\.1)$").columns.tolist()+cols
    # exceptions 
    cols_del=[c for c in cols_del if not c.endswith('0.1')]
    if drop_constants:
        df=df.rd.drop_constants()
    if not drop_unnamed:
        cols_del=[c for c in cols_del if not c.startswith('Unnamed')]
    if any(df.columns.duplicated()):
#         from roux.lib.set import unique
        if verb: logging.warning(f"duplicate column/s dropped:{df.loc[:,df.columns.duplicated()].columns.tolist()}")
        df=df.loc[:,~(df.columns.duplicated())]
    if len(cols_del)!=0:
        if verb: logging.warning(f"dropped columns: {', '.join(cols_del)}")
        return df.drop(cols_del,axis=1)
    else:
        return df
    
@to_rd
def compress(df1,coff_categories=20,test=False):
    """Compress the dataframe by converting columns containing strings/objects to categorical.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        coff_categories (int): if the number of unique values are less than cutoff the it will be converted to categories. 
        test (bool): verbose.
        
    Returns:
        df1 (DataFrame): output dataframe.
    """    
    if test: ini=df1.memory_usage().sum()
    ds=df1.select_dtypes('object').nunique()
    for c in ds[ds<=coff_categories].index:
        df1[c]=df1[c].astype('category')
    if test: logging.info(f"compression={((ini-df1.memory_usage().sum())/ini)*100:.1f}%")
    return df1

@to_rd
def clean_compress(df,kws_compress={},**kws_clean): 
    """`clean` and `compress` the dataframe.
    
    Parameters:
        df (DataFrame): input dataframe.
        kws_compress (int): keyword arguments for the `compress` function. 
        test (bool): verbose.
        
    Keyword Arguments:
        kws_clean (dict): parameters provided to `clean` function.
        
    Returns:
        df1 (DataFrame): output dataframe.

    See Also:
        `clean`
        `compress`
    """    
    return df.rd.clean(**kws_clean).rd.compress(**kws_compress)

## nans:    
@to_rd
def check_na(df,
             subset=None,
             perc=False,
            ):
    """Number/percentage of missing values in columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        perc (bool): output percentages.
        
    Returns:
        ds (Series): output stats.
    """    
    if subset is None: subset=df.columns.tolist()
    if not perc:
        return df.loc[:,subset].isnull().sum()
    else:
        return (df.loc[:,subset].isnull().sum()/df.loc[:,subset].agg(len))*100

@to_rd
def validate_no_na(df,subset=None):
    """Validate no missing values in columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        perc (bool): output percentages.
        
    Returns:
        ds (Series): output stats.
    """
    if subset is None: subset=df.columns.tolist()
    return not df.loc[:,subset].isnull().any().any()

@to_rd
def assert_no_na(df,subset=None):
    """Assert that no missing values in columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        perc (bool): output percentages.
        
    Returns:
        ds (Series): output stats.
    """    
    assert validate_no_na(df,subset=subset), check_na(df,subset=subset) 
    return df

## nunique:
@to_rd
def check_nunique(df,subset=None,groupby=None,perc=False):
    """Number/percentage of unique values in columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        perc (bool): output percentages.
        
    Returns:
        ds (Series): output stats.
    """
    if subset is None:
        subset=df.select_dtypes((object,bool)).columns.tolist()
    if isinstance(subset,str):
        subset=[subset]
    if groupby is None:
        if not perc:
            return df.loc[:,subset].nunique()
        else:
            return (df.loc[:,subset].nunique()/df.loc[:,subset].agg(len))*100
    else:
        from roux.lib.set import list2str
        return df.groupby(groupby)[list2str(subset)].nunique()

## nunique:
@to_rd
def check_inflation(df1,subset=None):
    """Occurances of values in columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        
    Returns:
        ds (Series): output stats.
    """
    if subset is None: subset=df1.columns.tolist()    
    if subset is None:
        subset=df1.columns.tolist()
    return df1.loc[:,subset].apply(lambda x: (x.value_counts().values[0]/len(df1))*100).sort_values(ascending=False)
    
## duplicates:
@to_rd
def check_dups(df,subset=None,perc=False):
    """Check duplicates.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        perc (bool): output percentages.
        
    Returns:
        ds (Series): output stats.
    """
    if subset is None: subset=df.columns.tolist()
    df1=df.loc[df.duplicated(subset=subset,keep=False),:].sort_values(by=subset)
    from roux.viz.annot import perc_label
    logging.info("duplicate rows: "+perc_label(len(df1),len(df)))
    if not perc:
        return df1
    else:
        return 100*(len(df1)/len(df))

@to_rd
def check_duplicated(df,subset=None,perc=False):
    """Check duplicates (alias of `check_dups`)    
    """
    return check_dups(df,subset=subset,perc=perc)

@to_rd
def validate_no_dups(df,subset=None,):
    """Validate that no duplicates.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
    """
    if subset is None: subset=df.columns.tolist()
    out=not df.duplicated(subset=subset).any()
    if not out: logging.warning('duplicate rows found')
    return out

@to_rd
def validate_no_duplicates(df,subset=None,):
    """Validate that no duplicates (alias of `validate_no_dups`)
    """
    return validate_no_dups(df,subset=subset,)

@to_rd
def assert_no_dups(df,subset=None):
    """Assert that no duplicates
    """    
    assert validate_no_dups(df,subset=subset), check_dups(df,subset=subset,perc=False)
    return df

## asserts        
@to_rd
def validate_dense(df01,subset=None,duplicates=True,na=True,message=None,):
    """Validate no missing values and no duplicates in the dataframe.
    
    Parameters:
        df01 (DataFrame): input dataframe.
        subset (list): list of columns.
        duplicates (bool): whether to check duplicates.
        na (bool): whether to check na.
        message (str): error message
        
    """    
    if subset is None:
        subset=df01.columns.tolist()
    if duplicates: df01.rd.validate_no_dups(subset=subset)#, 'duplicates found' if message is None else message
    if na: df01.rd.validate_no_na(subset=subset)# if message is None else message
    return True

@to_rd
def assert_dense(df01,subset=None,duplicates=True,na=True,message=None):
    """Alias of `validate_dense`.
    
    Notes:
        to be deprecated in future releases.
    """
    assert validate_dense(df01,subset=subset,duplicates=duplicates,na=na,message=message)
    return df01

## mappings
@to_rd
def classify_mappings(
    df1: pd.DataFrame,
    col1: str,
    col2: str,
    clean: bool=False,
    ) -> pd.DataFrame:
    """Classify mappings between items in two columns.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        col1 (str): column #1.
        col2 (str): column #2.
        clean (str): drop columns with the counts.
        
    Returns:
        (pd.DataFrame): output.
    """    
    df1[col2+' count']=df1.groupby(col1)[col2].transform('nunique')
    df1[col1+' count']=df1.groupby(col2)[col1].transform('nunique')
    df1['mapping']=df1.apply(lambda x: "1:1" if (x[col1+' count']==1) and (x[col2+' count']==1) else \
                                        "1:m" if (x[col1+' count']==1) else \
                                        "m:1" if (x[col2+' count']==1) else "m:m",
                                        axis=1)
    if clean:
        df1=df1.drop([col1+' count',col2+' count'],axis=1)
    return df1

@to_rd        
def check_mappings(df,
                   subset=None,
                   out='full',
                  ):
    """Mapping between items in two columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        out (str): format of the output.
        
    Returns:
        ds (Series): output stats.
    """
    if subset is None: subset=df.columns.tolist()
    import itertools
    d={}
    for t in list(itertools.permutations(subset,2)):
        d[t]=df.groupby(t[0])[t[1]].nunique().value_counts()
    df2=pd.concat(d,axis=0,ignore_index=False,names=['from','to','map to']).to_frame('map from').sort_index().reset_index(-1).loc[:,['map from','map to']]
    if out=='full':
        return df2
    else:
        return df2.loc[tuple(subset),:]#'map to'].item()

@to_rd        
def validate_1_1_mappings(df,
                   subset=None,
                  ):
    """Validate that the papping between items in two columns is 1:1.
    
    Parameters:
        df (DataFrame): input dataframe.
        subset (list): list of columns.
        out (str): format of the output.
        
    """
    df1=check_mappings(df,
                   subset=subset,
                  )
    assert all(df1['map to']==1), df1
    
@to_rd
def get_mappings(df1,
                 subset=None,
                 keep='1:1',
                 clean=False,
                 cols=None,
                ):
    """Classify the mapapping between items in two columns.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        subset (list): list of columns.
        keep (str): type of mapping (1:1|1:m|m:1).
        clean (bool): whether remove temporary columns. 
        cols (list): alias of `subset`.
        
    Returns:
        df (DataFrame): output dataframe.
    """
    if not cols is None and not subset is None: 
        logging.error(f"cols and subset are alias, both cannot be used.")
        return
    if cols is None and not subset is None: cols=subset        
    if cols is None: cols=df1.columns.tolist()
    if not df1.rd.validate_no_dups(cols):
        df1=df1.loc[:,cols].log.drop_duplicates()
    if len(cols)==2:
        from roux.lib.set import get_alt
        df2=df1.copy()
        cols2=[]
        for c in cols:
            d1=df2.groupby(c)[get_alt(cols,c)].nunique().to_dict()
            c2=f"{c}:{get_alt(cols,c)}"
            df2[c2]=df2[c].map(d1)
            cols2.append(c2)
        df2['mapping']=df2.loc[:,cols2].apply(lambda x: ':'.join(["1" if i==1 else 'm' for i in x]),axis=1)
        if keep=='1:1':
            df2=df2.rd.filter_rows({'mapping':'1:1'})
        else:
            logging.info(df2['mapping'].value_counts())
        if clean:
            df2=df2.drop(cols2+['mapping'],axis=1)
        return df2
    else:
        d1={'1:1':df1.copy(),
           }
        if keep!='1:1':
            d1['not']=pd.DataFrame()
        import itertools
    #     for t in list(itertools.permutations(cols,2)):
        for c in cols:
            d1['1:1']=d1['1:1'].groupby(c).filter(lambda df: len(df)==1)
            if keep!='1:1':
                d1['not']=d1['not'].append(df1.copy().groupby(c).filter(lambda df: len(df)!=1))
        if keep=='1:1':
            logging.info(df1.shape)
            logging.info(d1['1:1'].shape)
            return d1['1:1']
        else:
            assert(len(df1)==len(d1['1:1'])+len(d1['not']))
            return pd.concat(d1,axis=0,names=['mapping']).reset_index()


@to_rd
def groupby_filter_fast(df1,col,fun,how,coff):
    """Groupby and filter fast.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        col (str): column name.
        fun (object): function to filter with.
        how (str): greater or less than `coff` (>|<).  
        coff (float): cut-off.    
        
    Returns:
        df1 (DataFrame): output dataframe.
    
    Todo:
        Deprecation if `pandas.core.groupby.DataFrameGroupBy.filter` is faster.
    """
    ds1=df1.groupby(col).transform(fun)
    if how=='<':
        return df1.loc[(ds1<coff),:]
    elif how=='>':
        return df1.loc[(ds1>coff),:]
    else:
        raise ValueError(how)
        
@to_rd
def to_map_binary(df,colgroupby=None,colvalue=None):
    """Convert linear mappings to a binary map
    
    Parameters:
        df (DataFrame): input dataframe.
        colgroupby (str): name of the column for groupby.
        colvalue (str): name of the column containing values.
        
    Returns:
        df1 (DataFrame): output dataframe.
    """
    colgroupby=[colgroupby] if isinstance(colgroupby,str) else colgroupby
    colvalue=[colvalue] if isinstance(colvalue,str) else colvalue
    if not df.rd.validate_no_dups(colgroupby+colvalue):
        logging.warning('duplicates found')
        df=df.log.drop_duplicates(subset=colgroupby+colvalue)
    return (df
            .assign(_value=True)
            .pivot(index=colvalue,columns=colgroupby,values='_value')
            .fillna(False)
            )

## intersections 
@to_rd        
def check_intersections(df,
                        colindex=None, # 'samples'
                        colgroupby=None, # 'yticklabels'
                        plot=False,
                        **kws_plot):
    """Check intersections.
    Linear dataframe to is converted to a binary map and then to a series using `groupby`.

    Parameters:
        df (DataFrame): input dataframe.
        colindex (str): name of the index column.
        colgroupby (str): name of the groupby column.
        plot (bool): plot or not.
    
    Returns:
        ds1 (Series): output Series.

    Keyword Arguments:
        kws_plot (dict): parameters provided to the plotting function.
    """
    # if isinstance(colindex,str):
    #     colindex=[colindex]
    if isinstance(df,pd.DataFrame):
        if not (colgroupby is None or colindex is None) :
            if not all(df.dtypes==bool): 
#             if isinstance(colgroupby,str):
                # lin
                df1=to_map_binary(df,colgroupby=colgroupby,colvalue=colindex)
                ds=df1.groupby(df1.columns.to_list()).size()
            elif isinstance(colgroupby,(str,list)):
                assert(not df.rd.check_duplicated([colindex]+colgroupby))
                # map
                # df=df.set_index(colindex).loc[:,colgroupby] 
                # ds=df.groupby(df.columns.tolist()).size()
                ds=df.groupby(colgroupby).nunique(colindex)
            else:
                logging.error('colgroupby should be a str or list')
        else:
            # map
            ds=map2groupby(df)
    elif isinstance(df,pd.Series):
        ds=df
    # elif isinstance(df,dict):
    #     ds=dict2df(d1).rd.check_intersections(colindex='value',colgroupby='key')
    else:
        raise ValueError("data type of `df`")
    ds.name=colindex if isinstance(colindex, str) else ','.join(colindex) if isinstance(colindex, list) else None
    if plot:
        from roux.viz.bar import plot_intersections
        return plot_intersections(ds,**kws_plot)
    else:
        return ds

def get_totals(ds1):
    """Get totals from the output of `check_intersections`.
    
    Parameters:
        ds1 (Series): input Series.
    
    Returns:
        d (dict): output dictionary.
    """
    col=ds1.name if not ds1.name is None else 0
    df1=ds1.to_frame().reset_index()
    return {c:df1.loc[df1[c],col].sum() for c in ds1.index.names}
    
#filter df
@to_rd
def filter_rows(df,
                d,
                sign='==',
                logic='and',
                drop_constants=False,
                test=False,
                verb=True,
               ):
    """Filter rows using a dictionary.
    
    Parameters:
        df (DataFrame): input dataframe.
        d (dict): dictionary.
        sign (str): condition within mappings ('==').
        logic (str): condition between mappings ('and').
        drop_constants (bool): to drop the columns with single unique value (False).
        test (bool): testing (False).
        verb (bool): more verbose (True).
                
    Returns:
        df (DataFrame): output dataframe.
    
    """
    if verb: logging.info(df.shape)    
    assert(all([isinstance(d[k],(str,list)) for k in d]))
    qry = f" {logic} ".join([f"`{k}` {sign} "+(f"'{v}'" if isinstance(v,str) else f"{v}") for k,v in d.items()])
    df1=df.query(qry)
    if test:
        logging.info(df1.loc[:,list(d.keys())].drop_duplicates())
        logging.warning('may be some column names are wrong..')
        logging.warning([k for k in d if not k in df])
    if verb: logging.info(df1.shape)
    if drop_constants:
        df1=df1.rd.drop_constants()
    return df1

## conversion to type
@to_rd
def to_dict(df,cols,drop_duplicates=False):
    """DataFrame to dictionary.
    
    Parameters:
        df (DataFrame): input dataframe.
        cols (list): list of two columns: 1st contains keys and second contains value.
        drop_duplicates: whether to drop the duplicate values (False).
        
    Returns:
        d (dict): output dictionary.
    """
    df=df.log.dropna(subset=cols)
    if drop_duplicates:
        df=df.loc[:,cols].drop_duplicates()
    if not df[cols[0]].duplicated().any():
        return df.set_index(cols[0])[cols[1]].to_dict()
    else:
        logging.warning('format: {key:list}')
        assert df[cols[1]].dtype=='O', df[cols[1]].dtype
        return df.groupby(cols[0])[cols[1]].unique().to_dict()        

## to avoid overlap with `io_dict.to_dict`
del to_dict

## conversion
@to_rd
def get_bools(df,cols,drop=False):
    """Columns to bools. One-hot-encoder (`get_dummies`).
    
    Parameters:
        df (DataFrame): input dataframe. 
        cols (list): columns to encode.
        drop (bool): drop the `cols` (False).
    
    Returns: 
        df (DataFrame): output dataframe.
    """
    for c in cols:
        df_=pd.get_dummies(df[c],
                                  prefix=c,
                                  prefix_sep=": ",
                                  dummy_na=False)
        df_=df_.replace(1,True).replace(0,False)
        df=df.join(df_)
        if drop:
            df=df.drop([c],axis=1)
    return df

@to_rd
def agg_bools(df1,cols):
    """Bools to columns. Reverse of one-hot encoder (`get_dummies`). 
    
    Parameters:
        df1 (DataFrame): input dataframe.
        cols (list): columns. 
    
    Returns:
        ds (Series): output series.
    """
    col='+'.join(cols)
#     print(df1.loc[:,cols].T.sum())
    assert(all(df1.loc[:,cols].T.sum()==1))
    for c in cols:
        df1.loc[df1[c],col]=c
    return df1[col]     

## paired dfs
@to_rd
def melt_paired(df,
                cols_index=None, # paired
                suffixes=None,
                cols_value=None,
                ):
    """Melt a paired dataframe.
    
    Parameters:
        df (DataFrame): input dataframe.
        cols_index (list): paired index columns (None).
        suffixes (list): paired suffixes (None).
        cols_value (list): names of the columns containing the values (None).

    Notes:
        Partial melt melts selected columns `cols_value`.
    
    Examples:
        Paired parameters:
            cols_value=['value1','value2'],
            suffixes=['gene1','gene2'],
    """
    if cols_value is None:
        assert not (cols_index is None and suffixes is None), "either cols_index or suffixes needed" 
        if suffixes is None and not cols_index is None:
            from roux.lib.str import get_suffix
            suffixes=get_suffix(*cols_index,common=False, clean=True)
            
        # both suffixes should not be in any column name
        assert not any([all([s in c for s in suffixes]) for c in df]), "both suffixes should not be in a single column name"
        assert not any([c==s for s in suffixes for c in df]), "suffix should not be the column name"
        assert all([any([s in c for c in df]) for s in suffixes]), "both suffix should be in the column names"
        
        cols_common=[c for c in df if not any([s in c for s in suffixes])]
        dn2df={}
        for s in suffixes:
            cols=[c for c in df if s in c]
            dn2df[s]=df.loc[:,cols_common+cols].rename(columns={c:c.replace(s,'') for c in cols},
                                                      errors='raise')
        df1=pd.concat(dn2df,axis=0,names=['suffix']).reset_index(0)
        df2=df1.rename(columns={c: c[:-1] if c.endswith(' ') else c[1:] if c.startswith(' ') else c for c in df1},
                         errors='raise')
        if '' in df2:
            df2=df2.rename(columns={'':'id'},
                   errors='raise')
        assert len(df2)==len(df)*2
        return df2
    else:
        assert not suffixes is None
        import itertools
        df2=pd.concat({c: df.rename(columns={f"{c} {s}":f"value {s}" for s in suffixes},errors='raise') for c in cols_value},
                        axis=0,names=['variable'],
                        ).reset_index(0)
        if len(cols_value)>1:
            df2=df2.drop([f"{c} {s}" for c,s in itertools.product(cols_value,suffixes)],axis=1)
        assert len(df2)==len(df)*len(cols_value)
        return df2

@to_rd
def get_chunks(df1,colindex,colvalue,bins=None,value='right'):
    """Get chunks of a dataframe.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        colindex (str): name of the index column.
        colvalue (str): name of the column containing values [0-100]
        bins (int): number of bins.
        value (str): value to use as the name of the chunk ('right').
        
    Returns: 
        ds (Series): output series.
    """
    from roux.lib.set import unique,nunique
    if bins==0:
        df1['chunk']=bins
        logging.warning("bins=0, so chunks=1")
        return df1['chunk']
    elif bins is None:
        bins=int(np.ceil(df1.memory_usage().sum()/1e9))
    df2=df1.loc[:,[colindex,colvalue]].drop_duplicates()
    from roux.stat.transform import get_bins
    d1=get_bins(df2.set_index(colindex)[colvalue],
                 bins=bins,
                 value=value,
                ignore=True)
    ## number bins
    d_={k:f"chunk{ki+1:08d}_upto{int(k):03d}" for ki,k in enumerate(sorted(np.unique(list(d1.values()))))}
    ## rename bins
    d2={k:d_[d1[k]] for k in d1}
    assert(nunique(d1.values())==nunique(d2.values()))
    df1['chunk']=df1[colindex].map(d2)
    return df1['chunk']

## GROUPBY
# aggregate dataframes
def get_group(groups,i=None,verbose=True):
    """Get a dataframe for a group out of the `groupby` object.

    Parameters:
        groups (object): groupby object.
        i (int): index of the group (None).
        verbose (bool): verbose (True).
        
    Returns: 
        df (DataFrame): output dataframe.
        
    Notes: 
        Useful for testing `groupby`.
    """
    if not i is None: 
        dn=list(groups.groups.keys())[i]
    else:
        dn=groups.size().sort_values(ascending=False).index.tolist()[0]
    logging.info(dn)
    df=groups.get_group(dn)
    df.name=dn
    return df
        
# multiindex
def coltuples2str(cols,sep=' '):
    """Convert list of tuples (e.g. MultiIndex) to list of strings
    
    Parameters:
        cols (list): list of tuples.
        sep (str): separator within the joined tuples (' ').
        
    Returns:
        cols_str (list): list of strings.
    """
    from roux.lib.str import tuple2str
    cols_str=[]
    for col in cols:
        cols_str.append(tuple2str(col,sep=sep))
    return cols_str

@to_rd
def to_multiindex_columns(df,suffixes,test=False):
    """Single level columns to multiindex.
    
    Parameters:
        df (DataFrame): input dataframe.
        suffixes (list): list of suffixes.
        test (bool): verbose (False).
        
    Returns:
        df (DataFrame): output dataframe.
    """
    cols=[c for c in df if c.endswith(f' {suffixes[0]}') or c.endswith(f' {suffixes[1]}')]
    if test:
        logging.info(cols)
    df=df.loc[:,cols]
    df=df.rename(columns={c: (s,c.replace(f' {s}','')) for s in suffixes for c in df if c.endswith(f' {s}')},
                errors='raise')
    df.columns=pd.MultiIndex.from_tuples(df.columns)
    return df

## ranges
@to_rd
def to_ranges(df1,colindex,colbool,sort=True):
    """Ranges from boolean columns.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        colindex (str): column containing index items.
        colbool (str): column containing boolean values.
        sort (bool): sort the dataframe (True).
        
    Returns:
        df1 (DataFrame): output dataframe.
        
    TODO:
        compare with io_sets.bools2intervals.
    """
    import scipy as sc
    if sort:
        df1=df1.sort_values(by=colindex)
    df1['group']=sc.ndimage.measurements.label(df1[colbool].astype(int))[0]
    return df1.loc[(df1['group']!=0),:].groupby('group')[colindex].agg([min,max]).reset_index()

@to_rd
def to_boolean(df1):
    """Boolean from ranges.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        
    Returns:
        ds (Series): output series.
        
    TODO:
        compare with io_sets.bools2intervals.
    """    
    low, high = np.array(df1).T[:,:, None]
    a = np.arange(high.max() + 1)
    return ((a >= low) & (a <= high)).any(axis=0)


## sorting
def to_cat(ds1,cats,ordered = True):
    """To series containing categories.
    
    Parameters:
        ds1 (Series): input series.
        cats (list): categories.
        ordered (bool): if the categories are ordered (True).

    Returns:
        ds1 (Series): output series.
    """
    ds1=ds1.astype('category')
    ds1=ds1.cat.set_categories(new_categories = cats, ordered = ordered)
    assert(not ds1.isnull().any())
    return ds1

@to_rd
def sort_valuesby_list(df1,by,cats,**kws):
    """Sort dataframe by custom order of items in a column.
        
    Parameters:
        df1 (DataFrame): input dataframe.
        by (str): column.
        cats (list): ordered list of items.

    Keyword parameters:
        kws (dict): parameters provided to `sort_values`.
    
    Returns:
        df (DataFrame): output dataframe.
    """
    df1[by]=to_cat(df1[by],cats,ordered = True)
    return df1.sort_values(by=by, **kws)

## apply_agg
def agg_by_order(x,order):
    """Get first item in the order.

    Parameters: 
        x (list): list.
        order (list): desired order of the items.

    Returns:
        k: first item.
        
    Notes:
        Used for sorting strings. e.g. `damaging > other non-conserving > other conserving`

    TODO: 
        Convert categories to numbers and take min
    """
    if len(x)==1:
#         print(x.values)
        return list(x.values)[0]
    for k in order:
        if k in x.values:
            return k
def agg_by_order_counts(x,order):
    """Get the aggregated counts by order*.

    Parameters:
        x (list): list.
        order (list): desired order of the items.
        
    Returns:
        df (DataFrame): output dataframe.
    
    Examples:
        df=pd.DataFrame({'a1':['a','b','c','a','b','c','d'],
        'b1':['a1','a1','a1','b1','b1','b1','b1'],})
        df.groupby('b1').apply(lambda df : agg_by_order_counts(x=df['a1'],
                                                       order=['b','c','a'],
                                                       ))
    """    
    ds=x.value_counts()
    ds=ds.add_prefix(f"{x.name}=")
    ds[x.name]=agg_by_order(x,order)
    return ds.to_frame('').T

@to_rd
def groupby_sort_values(df,col_groupby,col_sortby,
                 subset=None,
                 col_subset=None,
                 func='mean',ascending=True):
    """Sort groups. 
    
    Parameters:
        df (DataFrame): input dataframe.
        col_groupby (str|list): column/s to groupby with.
        col_sortby (str|list): column/s to sort values with.
        subset (list): columns (None).
        col_subset (str): column containing the subset (None).
        func (str): aggregate function, provided to numpy ('mean').
        ascending (bool): sort values ascending (True).
        
    Returns:
        df (DataFrame): output dataframe.
    """
    gs=df.groupby(col_groupby)
    if subset is None:
        df1=gs.agg({col_sortby:getattr(np,func)}).reset_index()
        df2=df.merge(df1,
                on=col_groupby,how='inner',suffixes=['',f' per {col_groupby}'])
        logging.warning(f'column added to df: {col_sortby} per {col_groupby}')
        return df2.sort_values(f'{col_sortby} per {col_groupby}',ascending=ascending)
    else:
        df1=df.groupby(col_subset).get_group(subset)
        df2=df1.groupby(col_groupby).agg({col_sortby:getattr(np,func)}).reset_index()
        return sort_col_by_list(df, 
                                col_groupby,
                                df2.sort_values(col_sortby,
                                                ascending=ascending)[col_groupby])
#         return df2.sort_values(f'{col_sortby} per {col_groupby}',ascending=ascending)
sort_values_groupby=groupby_sort_values

@to_rd
def swap_paired_cols(df_,suffixes=['gene1', 'gene2']):
    """Swap suffixes of paired columns.
    
    Parameters:
        df_ (DataFrame): input dataframe.
        suffixes (list): suffixes.
    
    Returns: 
        df (DataFrame): output dataframe.    
    """
    rename={c:c.replace(suffixes[0],suffixes[1]) if (suffixes[0] in c) else c.replace(suffixes[1],suffixes[0]) if (suffixes[1] in c) else c for c in df_}
    return df_.rename(columns=rename,errors='raise')

@to_rd
def sort_columns_by_values(df,cols_sortby=['mutation gene1','mutation gene2'],
                            suffixes=['gene1','gene2'], # no spaces
                           clean=False,
                            ):
    """Sort the values in columns in ascending order.
    
    Parameters:
        df (DataFrame): input dataframe.
        cols_sortby (list): (['mutation gene1','mutation gene2'])
        suffixes (list): suffixes, without no spaces. (['gene1','gene2'])
        
    Returns:
        df (DataFrame): output dataframe.
        
    Notes:
        In the output dataframe, `sorted` means values are sorted because gene1>gene2.
    """
    df.rd.assert_no_na(subset=cols_sortby)
    suffixes=[s.replace(' ','') for s in suffixes]
    dn2df={}
    # keys: (equal, to be sorted)
    dn2df[(False,False)]=df.loc[(df[cols_sortby[0]]<df[cols_sortby[1]]),:]
    dn2df[(False,True)]=df.loc[(df[cols_sortby[0]]>df[cols_sortby[1]]),:]
    dn2df[(True,False)]=df.loc[(df[cols_sortby[0]]==df[cols_sortby[1]]),:]
    dn2df[(True,True)]=df.loc[(df[cols_sortby[0]]==df[cols_sortby[1]]),:]
    ## rename columns of of to be sorted
    ## TODO: use swap_paired_cols
    rename={c:c.replace(suffixes[0],suffixes[1]) if (suffixes[0] in c) else c.replace(suffixes[1],suffixes[0]) if (suffixes[1] in c) else c for c in df}
    for k in [True, False]:
        dn2df[(k,True)]=dn2df[(k,True)].rename(columns=rename,
                                              errors='raise')
        
    df1=pd.concat(dn2df,names=['equal','sorted']).reset_index([0,1])
    logging.info(df1.groupby(['equal','sorted']).size())
    if clean:
        df1=df1.drop(['equal','sorted'],axis=1)
    return df1
    
## ids
@to_rd
def make_ids(df,cols,ids_have_equal_length,sep='--',sort=False):
    """Make ids by joining string ids in more than one columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        cols (list): columns. 
        ids_have_equal_length (bool): ids have equal length, if True faster processing.
        sep (str): separator between the ids ('--').
        sort (bool): sort the ids before joining (False).
    
    Returns:
        ds (Series): output series.
    """
    get_ids=lambda x: '--'.join(x)
    get_ids_sorted=lambda x: '--'.join(sorted(x))
    if ids_have_equal_length:
        logging.warning("ids should be of equal character length and should not contain non-alphanumeric characters e.g. '.'")
        return np.apply_along_axis(get_ids if not sort else get_ids_sorted, 1, df.loc[:,cols].values)
    else:
        return df.loc[:,cols].agg(lambda x: sep.join(x if not sort else sorted(x)),axis=1)

@to_rd
def make_ids_sorted(df,cols,ids_have_equal_length,sep='--'):
    """Make sorted ids by joining string ids in more than one columns.
    
    Parameters:
        df (DataFrame): input dataframe.
        cols (list): columns. 
        ids_have_equal_length (bool): ids have equal length, if True faster processing.
        sep (str): separator between the ids ('--').
    
    Returns:
        ds (Series): output series.
    """    
    return make_ids(df,cols,ids_have_equal_length,sep=sep,sort=True)
    
def get_alt_id(s1='A--B',s2='A',sep='--'): 
    """Get alternate/partner id from a paired id.
    
    Parameters:
        s1 (str): joined id.
        s2 (str): query id. 
        
    Returns:
        s (str): partner id.
    """    
    return [s for s in s1.split(sep) if s!=s2][0]

@to_rd    
def split_ids(df1,col,sep='--',prefix=None):
    """Split joined ids to individual ones.
    
    Parameters:
        df1 (DataFrame): input dataframe.
        col (str): column containing the joined ids.
        sep (str): separator within the joined ids ('--').
        prefix (str): prefix of the individual ids (None).
    
    Return:
        df1 (DataFrame): output dataframe.
    """
    # assert not df1._is_view, "input series should be a copy not a view"
    df=df1[col].str.split(sep,expand=True)
    for i in range(len(df.columns)):
        df1[f"{col} {i+1}"]=df[i].copy()
    if not prefix is None:
        df1=df1.rd.renameby_replace(replaces={f"{col} ":prefix})
    return df1

## tables io
def dict2df(d,colkey='key',colvalue='value'):
    """Dictionary to DataFrame.
    
    Parameters:
        d (dict): dictionary.
        colkey (str): name of column containing the keys.
        colvalue (str): name of column containing the values.
    
    Returns:
        df (DataFrame): output dataframe.
    """
    if not isinstance(list(d.values())[0],list):
        return pd.DataFrame({colkey:d.keys(), colvalue:d.values()})
    else:
        return pd.DataFrame(pd.concat({k:pd.Series(d[k]) for k in d})).droplevel(1).reset_index().rename(columns={'index':colkey,0:colvalue},
                                                                                                        errors='raise')
def log_shape_change(d1,fun=''):
    """Report the changes in the shapes of a DataFrame.

    Parameters:
        d1 (dic): dictionary containing the shapes.
        fun (str): name of the function.
    """
    if d1['from']!=d1['to']:
        prefix=f"{fun}: " if fun!='' else ''
        if d1['from'][0]==d1['to'][0]:
            logging.info(f"{prefix}shape changed: {d1['from']}->{d1['to']}, length constant")
        elif d1['from'][1]==d1['to'][1]:
            logging.info(f"{prefix}shape changed: {d1['from']}->{d1['to']}, width constant")        
        else:
            logging.info(f"{prefix}shape changed: {d1['from']}->{d1['to']}")
## log
def log_apply(df, fun, 
              validate_equal_length=False,
              validate_equal_width=False,
              validate_equal_shape=False,
              validate_no_decrease_length=False,
              validate_no_decrease_width=False,
              validate_no_increase_length=False,
              validate_no_increase_width=False,
              *args, **kwargs):
    """Report (log) the changes in the shapes of the dataframe before and after an operation/s.
    
    Parameters:
        df (DataFrame): input dataframe.
        fun (object): function to apply on the dataframe.
        validate_equal_length (bool): Validate that the number of rows i.e. length of the dataframe remains the same before and after the operation. 
        validate_equal_width (bool): Validate that the number of columns i.e. width of the dataframe remains the same before and after the operation. 
        validate_equal_shape (bool): Validate that the number of rows and columns i.e. shape of the dataframe remains the same before and after the operation. 
    
    Keyword parameters:
        args (tuple): provided to `fun`.
        kwargs (dict): provided to `fun`.
    
    Returns:
        df (DataFrame): output dataframe.
    """
    d1={}
    d1['from']=df.shape
    if isinstance(fun,str):
        df = getattr(df, fun)(*args, **kwargs)
    else:
        df = fun(df,*args, **kwargs)
    d1['to']=df.shape
    log_shape_change(d1,fun=fun)
    if validate_equal_length: assert d1["from"][0]==d1["to"][0], (d1["from"][0],d1["to"][0])
    if validate_equal_width: assert d1["from"][1]==d1["to"][1], (d1["from"][1],d1["to"][1])
    if validate_no_decrease_length: assert d1["from"][0]<=d1["to"][0], (d1["from"][0],d1["to"][0])
    if validate_no_decrease_width: assert d1["from"][1]<=d1["to"][1], (d1["from"][1],d1["to"][1])
    if validate_no_increase_length: assert d1["from"][0]>=d1["to"][0], (d1["from"][0],d1["to"][0])
    if validate_no_increase_width: assert d1["from"][1]>=d1["to"][1], (d1["from"][1],d1["to"][1])
    if validate_equal_shape: assert d1["from"]==d1["to"], (d1["from"],d1["to"])
    return df

@pd.api.extensions.register_dataframe_accessor("log")
class log:
    """Report (log) the changes in the shapes of the dataframe before and after an operation/s.
        
    TODO:
        Create the attribures (`attr`) using strings e.g. setattr.
        import inspect
        fun=inspect.currentframe().f_code.co_name
    """    
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
    def __call__(self,col=None):
        if not col is None and col in self._obj:
            logging.info(f"nunique('{col}') = {self._obj[col].nunique()}")
        logging.info(f"shape = {self._obj.shape}")
        return self._obj
    def dropna(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='dropna',**kws)
    def drop_duplicates(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='drop_duplicates',**kws)    
    def drop(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='drop',**kws)    
    def query(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='query',**kws)    
    def filter_(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='filter',**kws)    
    def pivot(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='pivot',**kws)
    def pivot_table(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='pivot_table',**kws)    
    def melt(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='melt',**kws)
    def stack(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='stack',**kws)
    def unstack(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='unstack',**kws)
    def explode(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='explode',**kws)
    def merge(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='merge',**kws)
    def join(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='join',**kws)    
    def groupby(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun='groupby',**kws)
    ## rd
    def clean(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun=clean,**kws)
    def filter_rows(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun=filter_rows,**kws)
    def melt_paired(self,**kws):
        from roux.lib.df import log_apply
        return log_apply(self._obj,fun=melt_paired,**kws)
    ## rd/logging-only
    def check_nunique(self,**kws):
        logging.info(f'unique {kws}')
        logging.info(check_nunique(self._obj,**kws))
        return self._obj    
    def check_na(self,**kws):
        logging.info(f'na {kws}')
        logging.info(check_na(self._obj,**kws))
        return self._obj    
    def check_dups(self,**kws):
        logging.info(f'dups {kws}')
        logging.info(check_dups(self._obj,**kws))
        return self._obj        


"""For processing multiple pandas DataFrames/Series"""
from roux.lib.df import *
from roux.lib import to_rd
        
def filter_dfs(dfs,cols,how='inner'):
    """Filter dataframes based items in the common columns.

    Parameters:
        dfs (list): list of dataframes.
        cols (list): list of columns.
        how (str): how to filter ('inner')
    
    Returns
        dfs (list): list of dataframes.        
    """
    def apply_(dfs,col,how):
        from roux.lib.set import list2intersection,list2union
        if how=='inner':
            l=list(list2intersection([df[col].tolist() for df in dfs]))
        elif how=='outer':
            l=list(list2union([df[col].tolist() for df in dfs]))
        else:
            raise ValueError("how")
        logging.info(f"len({col})={len(l)}")
        return [df.loc[(df[col].isin(l)),:] for df in dfs]
    if isinstance(cols,str):
        cols=[cols]
    # sort columns by nunique
    cols=dfs[0].loc[:,cols].nunique().sort_values().index.tolist()
    for c in cols:
        dfs=apply_(dfs=dfs,col=c,how=how)
    return dfs

@to_rd
def merge_paired(
    df1,df2,
    left_ons, # suffixed
    right_on, # to be suffixed
    common=[], # not suffixed
    right_ons_common=[], # not to be suffixed
    how='inner',
    validates=['1:1','1:1'],
    suffixes=None,
    test=False,
    verb=True,
    **kws,
    ):
    """Merge uppaired dataframes to a paired dataframe. 
    
    Parameters:
        df1 (DataFrame): paired dataframe.  
        df2 (DataFrame): unpaired dataframe.
        left_ons (list): columns of the `df1` (suffixed).
        right_on (str|list): column/s of the `df2` (to be suffixed).
        common (str|list): common column/s between `df1` and `df2` (not suffixed).
        right_ons_common (str|list): common column/s between `df2` to be used for merging (not to be suffixed).
        how (str): method of merging ('inner').
        validates (list): validate mappings for the 1st mapping between `df1` and `df2` and 2nd one between `df1+df2` and `df2` (['1:1','1:1']).
        suffixes (list): suffixes to be used (None).
        test (bool): testing (False).
        verb (bool): verbose (True).
     
    Keyword Parameters:
        kws (dict): parameters provided to `merge`.
    
    Returns:
        df (DataFrame): output dataframe.
    
    Examples:
        Parameters:
            how='inner',
            left_ons=['gene id gene1','gene id gene2'], # suffixed
            common='sample id', # not suffixed
            right_on='gene id', # to be suffixed
            right_ons_common=[], # not to be suffixed    
    """
    if isinstance(right_on,str):
        right_on=[right_on]
    if isinstance(right_ons_common,str):
        right_ons_common=[right_ons_common]
    if isinstance(common,str):
        common=[common]
        
    if isinstance(left_ons[0],list):
        logging.erorr(f'groupby on common index instead')
        return 
    if suffixes is None:
        from roux.lib.str import get_suffix
        suffixes=get_suffix(*left_ons,common=False, clean=True)
        suffixes=[f" {s}" for s in suffixes]
    d1={}
    d1['from']=df1.shape
    
    def apply_(df2,cols_on,suffix,test=False):
        if len(cols_on)!=0:
            df2=df2.set_index(cols_on)
        df2=df2.add_suffix(suffix)
        if len(cols_on)!=0:
            df2=df2.reset_index()
        if test: print(df2.columns)
        return df2
    df3=df1.copy()
    for i,(suffix,validate) in enumerate(zip(suffixes,validates)):
        if test:
            print(df3.columns.tolist())
            print([f"{s}{suffix}" for s in right_on]+right_ons_common)
        df3=df3.merge(
        right=apply_(df2,common+right_ons_common,suffix,test=test),
        on=[f"{s}{suffix}" for s in right_on]+common+(right_ons_common if i==1 else []),
        how=how,
        validate=validate,
        **kws,
        )
    d1['to']=df3.shape        
    if verb:log_shape_change(d1)        
    return df3

## append

## merge dfs
def merge_dfs(dfs,
             **kws):
    """Merge dataframes from left to right.   
    
    Parameters:
        dfs (list): list of dataframes.
        
    Keyword Parameters:
        kws (dict): parameters provided to `merge`.
    
    Returns:
        df (DataFrame): output dataframe.
        
    Notes:
        For example, reduce(lambda x, y: x.merge(y), [1, 2, 3, 4, 5]) merges ((((1.merge(2)).merge(3)).merge(4)).merge(5)). 
    """ 
    if kws['on']!='outer': logging.warning("Drop-outs may occur if on!='outer'. Make sure that the dataframes are ordered properly from left to right.")
    from functools import reduce
    logging.info(f"merge_dfs: shape changed from : dfs shape={[df.shape for df in dfs]}")
    df3=reduce(lambda df1,df2: pd.merge(df1,df2,**kws), dfs)
    logging.info(f"merge_dfs: shape changed to   : {df3.shape}")
    return df3

def compare_rows(df1,df2,
                 test=False,
                 **kws,):
    cols=list(set(df1.columns.tolist()) & set(df1.columns.tolist()))
    if test: info(cols)
    cols_sort=list(set(df1.select_dtypes(object).columns.tolist()) & set(cols))
    if test: info(cols_sort)
    if len(df1)==len(df2):
        return df1.loc[:,cols].sort_values(cols_sort).reset_index(drop=True).compare(
        df2.loc[:,cols].sort_values(cols_sort).reset_index(drop=True),
        keep_equal=True,
            **kws
        )#.rd.assert_no_na()
    else:
        logging.warning(f'unequal lengths: {len(df1)}!={len(df2)}')
        df_=df1.loc[:,cols].merge(right=df1.loc[:,cols],
                                    on=cols_sort,
                                    how='outer',
                                    indicator=True)
        logging.info(df_['_merge'].value_counts())
        return df_.loc[(df_['_merge']!='both'),:]


"""Sets."""
import itertools
import numpy as np
import pandas as pd
import logging

## io
def to_list(l1,p):
    """Save list.
    
    Parameters:
        l1 (list): input list.
        p (str): path.
    
    Returns:
        p (str): path.        
    """
    from roux.lib.sys import makedirs
    if not 'My Drive' in p:
        p=p.replace(' ','_')
    else:
        logging.warning('probably working on google drive; space/s left in the path.')
    makedirs(p)    
    with open(p,'w') as f:
        f.write('\n'.join(l1))
    return p

def read_list(p):
    """Read the lines in the file.

    Args:
        p (str): path.

    Returns:
        l (list): list.
    """
    with open(p,'r') as f:
        s=f.readlines()
    return s
# alias to be deprecated
read_lines=read_list

from functools import reduce
def union(l):
    """Union of lists.
    
    Parameters: 
        l (list): list of lists.
        
    Returns:
        l (list): list.
    """
    return reduce(np.union1d, (l))
def intersection(l):
    """Intersections of lists.
    
    Parameters: 
        l (list): list of lists.
        
    Returns:
        l (list): list.
    """
    return reduce(np.intersect1d, (l))

# aliases: to be deprecated in th future 
list2union=union
list2intersection=intersection

def nunion(l):
    """Count the items in union.
    
    Parameters: 
        l (list): list of lists.
        
    Returns:
        i (int): count.    
    """
    return len(union(l))

def nintersection(l):
    """Count the items in intersetion.
    
    Parameters: 
        l (list): list of lists.
        
    Returns:
        i (int): count.    
    """
    return len(intersection(l))

# lists mostly for agg
def dropna(x):
    """Drop `np.nan` items from a list.
    
    Parameters:
        x (list): list.
    
    Returns:
        x (list): list.
    """
    x_=[]
    for i in x:
        if not pd.isnull(i):
            x_.append(i)
    return x_

def unique(l):
    """Unique items in a list.
    
    Parameters:
        l (list): input list.
    
    Returns:
        l (list): list.
    """
    return list(np.unique(l))
    
def list2str(x,ignore=False):
    """Returns string if single item in a list.
    
    Parameters:
        x (list): list
        
    Returns:
        s (str): string.        
    """
    x=list(x)
    if not ignore:
        assert(len(x)==1)        
    if len(x)>1:
        logging.warning('more than 1 str value encountered, returning list')
        return x
    else:
        return x[0]

def unique_str(l,**kws):
    """Unique single item from a list.
    
    Parameters:
        l (list): input list.
    
    Returns:
        l (list): list.
    """
    return list2str(unique(dropna(l)))    

def nunique(l,**kws):
    """Count unique items in a list
    
    Parameters:
        l (list): list
    
    Returns:
        i (int): count.
    """
    return len(unique(l,**kws))

def flatten(l):
    """List of lists to list.
    
    Parameters:
        l (list): input list.
        
    Returns:
        l (list): output list.
    """    
    return list(np.hstack(np.array(l,dtype=object)))

def get_alt(l1,s,): 
    """Get alternate item between two.
        
    Parameters:
        l1 (list): list.
        s (str): item.
        
    Returns: 
        s (str): alternate item.
    """
    assert(s in l1)
    return [i for i in l1 if i!=s][0]

def jaccard_index(l1,l2):
    # if len(l1)==0 or len(l2)==0:
    #     return 0,0,0
    x1=len(set(l1).intersection(set(l2)))
    x2=len(set(l1).union(set(l2)))
    if x1==0 or len(l1)==0 or len(l2)==0:
        return 0,x1,x2
    else:
        return x1/x2,x1,x2
    
def intersections(dn2list,jaccard=False,count=True,fast=False,test=False):
    """Get intersections between lists.
    
    Parameters:
        dn2list (dist): dictionary mapping to lists.
        jaccard (bool): return jaccard indices.
        count (bool): return counts.
        fast (bool): fast.
        test (bool): verbose.
    
    Returns:
        df (DataFrame): output dataframe.
            
    TODOs: 
        1. feed as an estimator to `df.corr()`.
        2. faster processing by filling up the symetric half of the adjacency matrix.
    """
    dn2list={k:dropna(dn2list[k]) for k in dn2list}
    df=pd.DataFrame(index=dn2list.keys(),
                columns=dn2list.keys())
    if jaccard:
        dn2list={k:set(dn2list[k]) for k in dn2list}
    from tqdm import tqdm
    for k1i,k1 in tqdm(enumerate(dn2list.keys())):
#         if test:
#             print(f"{(k1i/len(dn2list.keys()))*100:.02f}")
        for k2i,k2 in enumerate(dn2list.keys()):
            if fast and k1i>=k2i:
                continue
            if jaccard:
                if len(dn2list[k1].union(dn2list[k2]))!=0:
                    l=jaccard_index(dn2list[k1],dn2list[k2])
                    # l=len(set(dn2list[k1]).intersection(dn2list[k2]))/len(dn2list[k1].union(dn2list[k2]))
                else:
                    l=np.nan
            else:
                l=list(set(dn2list[k1]).intersection(dn2list[k2]))
            if count:
                df.loc[k1,k2]=len(l)
            else:
                df.loc[k1,k2]=l
    return df

## ranges
def range_overlap(l1,l2):
    """Overlap between ranges.
    
    Parameters:
        l1 (list): start and end integers of one range.
        l2 (list): start and end integers of other range.
        
    Returns:
        l (list): overlapped range. 
    """
    return list(set.intersection(set(range(l1[0],l1[1]+1,1)),
                            set(range(l2[0],l2[1]+1,1))))

from roux.lib.df import to_ranges,to_boolean
def get_windows(a, 
                size = None, 
                overlap = None, 
                windows=None, 
                overlap_fraction=None,
                stretch_last=False,
                out_ranges=True,
               ):
    """Windows/segments from a range. 
    
    Parameters:
        a (list): range.
        size (int): size of the windows. 
        windows (int): number of windows.
        overlap_fraction (float): overlap fraction. 
        overlap (int): overlap length. 
        stretch_last (bool): stretch last window. 
        out_ranges (bool): whether to output ranges.
    
    Returns:
        df1 (DataFrame): output dataframe.
        
    Notes:
        1. For development, use of `int` provides `np.floor`.
    """
    if not windows is None and size is None:
        # TODOs
        size = int(len(a)/windows)
    if not overlap_fraction is None and overlap is None:
        overlap=int(size*overlap_fraction)
    shape = (a.size - size + 1, size)
    strides = a.strides * 2
    a2=np.lib.stride_tricks.as_strided(a, strides = strides, shape = shape)[0::overlap]
    if not out_ranges:
        df1=pd.DataFrame(a2.T)#.add_prefix('window#')
        df1.index.name='position in window'
        df1=df1.melt(ignore_index=False,value_name='position',var_name='window#').reset_index()
        if stretch_last:
            if df1.iloc[-1,:]['position']!=a[-1]:
                a3=range(df1.iloc[-1,:]['position']+1,a[-1]+1)
                df_=pd.DataFrame({
                                  'position':np.repeat(df1.iloc[-1,:]['position'],len(a3)),
                                  'window#':np.repeat(df1.iloc[-1,:]['window#'],len(a3)),
                                  'position':a3,})
                df1=pd.concat([df1,df_],axis=0)
        return df1
    else:
        df1=pd.DataFrame(
                    {'window#':range(a2.shape[0]),
                      'start':a2.min(axis=1),
                      'end':a2.max(axis=1),
                     },
                )
        if stretch_last:
            if df1.iloc[-1,:]['end']!=a[-1]:
                df1.iloc[-1,:]['end']=a[-1]
        return df1
    
def bools2intervals(v):
    """Convert bools to intervals.
    
    Parameters:
        v (list): list of bools.
        
    Returns:
        l (list): intervals.
    """
    return np.flatnonzero(np.diff(np.r_[0,v,0])!=0).reshape(-1,2) - [0,1]

def list2ranges(l):    
    ls=[]
    for l in zip(l[:-1],l[1:]):
        ls.append(l)
    return ls

# def dfbool2intervals(df,colbool):
#     """
#     ds contains bool values
#     """
#     df.index=range(len(df))
#     intervals=bools2intervals(df[colbool])
#     for intervali,interval in enumerate(intervals):
#         df.loc[interval[0]:interval[1],f'{colbool} interval id']=intervali
#         df.loc[interval[0]:interval[1],f'{colbool} interval start']=interval[0]
#         df.loc[interval[0]:interval[1],f'{colbool} interval stop']=interval[1]
#         df.loc[interval[0]:interval[1],f'{colbool} interval length']=interval[1]-interval[0]+1
#         df.loc[interval[0]:interval[1],f'{colbool} interval within index']=range(interval[1]-interval[0]+1)    
#     df[f'{colbool} interval index']=df.index    
#     return df

def get_pairs(
    items: list,
    items_with: list = None,
    size: int = 2,
    with_self: bool =False, # itertools.combinations does not pair self
    ) -> pd.DataFrame:
    """
    Creates a dataframe with the paired items.

    Parameters:
        items: the list of items to pair.
	items_with: list of items to pair with.
	size: size of the combinations.
	with_self: pair with self or not.

    Returns:
	table with pairs of items.

    Notes:
        1. the ids of the items are sorted e.g. 'a'-'b' not 'b'-'a'.
    """
    ## get lists
    items=list(set(items)) ## unique, sorted
    if items_with is None:
        items_with=[]
    else:
        items_with=list(set(items_with))
    ## arrage
    if len(items_with)==0:
        o1=itertools.combinations(items,size)
    else:
        # assert len(set(items) & set(items_with))==0, f'two lists should be non-overlapping, otherwise pairs with self would be created. {set(items) & set(items_with)}'
        o1=itertools.product(items,items_with)
        if not with_self and len(set(items) & set(items_with))!=0:
            o1=[(k1,k2) for k1,k2 in o1 if k1!=k2]
    # create dataframe
    df0=pd.DataFrame(o1,columns=range(1,size+1))
    return df0



from roux.global_imports import *   

# sequence
def plot_domain(
    d1: dict,
    x: float=1,
    xoff: float=0,
    y: float=1,
    height: float=0.8,
    ax: plt.Axes=None,
    **kws,
    ) -> plt.Axes:
    """Plot protein domain.

    Args:
        d1 (dict): plotting data including intervals.
        x (float, optional): x position. Defaults to 1.
        xoff (float, optional): x-offset. Defaults to 0.
        y (float, optional): y position. Defaults to 1.
        height (float, optional): height. Defaults to 0.8.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if ax is None:
        fig,ax=plt.subplots(figsize=[3,3])
    ## plot (start -> domain start) and (domain end -> end) separately
#     ax.plot([x+xoff,d1['start']+xoff],[y,y],color='k',zorder=1)
#     ax.plot([d1['end']+xoff,d1['protein length']+xoff],[y,y],color='k',zorder=1)
#     print(d1)
#     print([x+xoff,d1['start']+xoff])
#     print([d1['end']+xoff,d1['protein length']+xoff])

    ax.plot([x+xoff,d1['protein length']+xoff],[y,y],color='k',zorder=1)
    if pd.isnull(d1['type']):
        return ax
    patches=[]
    import matplotlib.patches as mpatches
    width=d1['end']-d1['start']
    patches.append(mpatches.Rectangle(
        xy=(d1['start']+xoff,
            y-(height*0.5)),
        width=width,
        height=height,
#         color=color,
#         mutation_scale=10,
#     #     mutation_aspect=1.5,
        joinstyle='round',
#         fc='none',
        zorder=2,
        **kws,
    ))
    _=[ax.add_patch(p) for p in patches]
    return ax

def plot_protein(
    df: pd.DataFrame,
    ax: plt.Axes=None,
    label: str=None,
    alignby: str=None,
    test: bool=False,
    **kws
    ) -> plt.Axes:
    """Plot protein.

    Args:
        df (pd.DataFrame): input data.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        label (str, optional): proein name. Defaults to None.
        alignby (str, optional): align proteins by this domain. Defaults to None.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if ax is None:
        fig,ax=plt.subplots(figsize=[3,3])
    if len(df)==1 and pd.isnull(df.iloc[0,:]['type']):
        alignby=None
    if not alignby is None:
        if alignby in df['description'].tolist():
            xoff=df.loc[(df['description']==alignby),:].iloc[0,:]['start']*-1
        else:
            xoff=0
    else:
        xoff=df.sort_values(by=['start']).iloc[0,:]['start']*-1
    _=df.apply(lambda x: plot_domain(d1=x,y=x['y'],
                                        ax=ax,
                                        xoff=xoff,
                                        color=None if not 'color' in x else x['color'],
                                     label=x['description'],
                                        **kws),axis=1)        
    if not label is None:
        ax.text(-10+xoff,df['y'].tolist()[0],label,ha='right',va='center')
#         ax.text(0,df['y'].tolist()[0],label,ha='left',va='bottom')
    return ax

def plot_gene(
    df1: pd.DataFrame,
    label: str=None,
    kws_plot: dict={},
    test: bool=False,
    outd: str=None,
    ax: plt.Axes=None,
    off_figw: float=1,
    off_figh: float=1,
    #**kws_plot_protein
    ) -> plt.Axes:
    """Plot genes.

    Args:
        df1 (pd.DataFrame): input data.
        label (str, optional): label to show. Defaults to None.
        kws_plot (dict, optional): parameters provided to the `plot` function. Defaults to {}.
        test (bool, optional): test mode. Defaults to False.
        outd (str, optional): output directory. Defaults to None.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        off_figw (float, optional): width offset. Defaults to 1.
        off_figh (float, optional): height offset. Defaults to 1.

    Returns:
        plt.Axes: `plt.Axes` object.
    """

    if hasattr(df1,'name'): 
        geneid=df1.name
    else:
        geneid=df1.iloc[0,:]['gene id']
    if 'title' in df1:
        kws_plot['title']=df1.iloc[0,:]['title']
    elif 'gene symbol' in df1:
        kws_plot['title']=df1.iloc[0,:]['gene symbol']
    else:
        kws_plot['title']=geneid
    if ax is None:
        figsize=[((df1['protein length'].max()-75)/250)*off_figw,
                                 ((df1['protein id'].nunique()+2)*0.3)*off_figh]
        fig,ax=plt.subplots(figsize=figsize)
    if label=='index':
        df1['yticklabel']=df1['y']*-1
    if df1['description'].isnull().all():
        alignby=None
    else:
        alignby=df1['description'].value_counts().index[0]
    _=df1.groupby('protein id',
                 sort=False).apply(lambda df1: plot_protein(df1,
                                                           label=df1.iloc[0,:]['yticklabel'] if (('yticklabel' in df1) and (not label is None)) else df1.name if (not label is None) else label,
                                                           alignby=alignby,
                                                           ax=ax,
                                                           test=False))
    if not test:ax.axis('off')
    ax.set(**{k:v for k,v in kws_plot.items() if not k=='title'},)
    set_label(0.05,1,f"${kws_plot['title']}$",ax=ax,va='bottom')
    return ax

def plot_genes_legend(df: pd.DataFrame,d1: dict):
    """Make the legends for the genes.

    Args:
        df (pd.DataFrame): input data.
        d1 (dict): plotting data.
    """
    fig,ax=plt.subplots()
#     d1=df.set_index('description')['color'].dropna().drop_duplicates().to_dict()
    import matplotlib.patches as mpatches
    l1=[mpatches.Patch(color=d1[k], label=k) for k in d1]        
    savelegend(f"plot/schem_gene_{' '.join(df['gene id'].unique()[:5]).replace('.',' ')}_legend.png",
           legend=plt.legend(handles=l1,frameon=True,title='domains/regions'),
           )
#     df.loc[df['color'].isnull(),'color']=(0,0,0,1)
    df['color']=df['color'].fillna('k').apply(str)
    # to_table(df,f"plot/schem_gene_{' '.join(df['gene id'].unique()[:5]).replace('.',' ')}_legend.pqt")

from roux.query.ensembl import to_protein_seq,to_domains
def plot_genes_data(
    df1: pd.DataFrame,
    release: int,
    species: str,
    custom: bool=False,
    colsort: str=None,
    cmap: str='Spectral',
    fast: bool=False
    ) -> tuple:
    """Plot gene-wise data.

    Args:
        df1 (pd.DataFrame): input data.
        release (int): Ensembl release.
        species (str): species name.
        custom (bool, optional): customised. Defaults to False.
        colsort (str, optional): column to sort by. Defaults to None.
        cmap (str, optional): colormap. Defaults to 'Spectral'.
        fast (bool, optional): parallel processing. Defaults to False.

    Returns:
        tuple: (dataframe, dictionary)
    """
    species=species.lower().replace(' ','_')
    if not custom:
        from pyensembl import EnsemblRelease
        ensembl=EnsemblRelease(release=release,species=species)
        if not 'protein length' in df1:
            df1['protein length']=df1['protein id'].progress_apply(lambda x: to_prtseq(x,ensembl,length=True))
        from roux.query.ensembl import gid2gname
        df1['gene symbol']=df1['gene id'].apply(lambda x: gid2gname(x,ensembl))
        # getattr(df1,'parallel_apply' if fast else "apply")
        df2=getattr(df1.groupby(['gene id','protein id']),("progress" if not (fast or len(df1<20)) else 'parallel')+"_apply")(lambda df: proteinid2domains(get_name(df,'protein id'),
        species=species,release=release)).reset_index().rd.clean()
        df2=df2.log.dropna(subset=['description'])
        df2['description']=df2['description'].replace('',np.nan).fillna(df2['id'])
        if len(df2)!=0:
            df2=df1.merge(df2,
                         on=['gene id','protein id'],
                         how='left',
                         )
        else:
            df2=df1.copy()
            for c in ['description','type']: 
                df2[c]=np.nan
#     df2=df2.log.dropna(subset=['type'])
    else:
        df2=df1.copy()

    df2['start']=df2.apply(lambda x: 1 if pd.isnull(x['type']) else x['start'],axis=1)
    df2['end']=df2.apply(lambda x: x['protein length'] if pd.isnull(x['type']) else x['end'],axis=1)
    #     print(df2.columns)
    df2['domains']=df2['protein id'].map(df2.dropna(subset=['description']).groupby('protein id').size().to_dict())
    df2['domains']=df2['domains'].fillna(0)
    if colsort is None:
        df2=df2.sort_values(['gene id','domains','protein length','start'],ascending=[True,False,False,True])
    else:
        df2=df2.sort_values(['gene id',colsort],ascending=[True,True])        
    def gety(df):
        df['y']=(df.groupby('protein id',sort=False).ngroup()+1)*-1
        return df
    df2=df2.groupby('gene id',as_index=False).apply(gety)
#     df=df.sort_values(by=['domains','protein length'],ascending=[False,False])        
    
    from roux.viz.colors import get_ncolors
    cs=get_ncolors(n=df2['description'].nunique(),
                   cmap=cmap, ceil=False,
                   vmin=0.2,vmax=1)
    d1=dict(zip(df2['description'].dropna().unique(),
                cs
               ))
    df2['color']=df2['description'].map(d1)
    return df2,d1
def plot_genes(df1,
    custom=False,
    colsort=None,
    release=100,
    cmap='Spectral',
    **kws_plot_gene
    ):
    """Plot many genes.

    Args:
        df1 (pd.DataFrame): input data.
        release (int): Ensembl release.
        custom (bool, optional): customised. Defaults to False.
        colsort (str, optional): column to sort by. Defaults to None.
        cmap (str, optional): colormap. Defaults to 'Spectral'.
    
    Keyword Args:
        kws_plot_gene: parameters provided to the `plot_genes_data` function. 
    
    Returns:
        tuple: (dataframe, dictionary)
    """
    df2,d1=plot_genes_data(df1,custom=custom,colsort=colsort,
                           release=release,
                           cmap=cmap)
    axs=df2.groupby('gene id').apply(plot_gene,**kws_plot_gene)
    plot_genes_legend(df2,d1)
    return axs,df2    